/* GDC -- D front-end for GCC
   Copyright (C) 2004 David Friedman
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
 
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
 
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "d-gcc-includes.h"
#include "d-lang.h"
#include "d-codegen.h"
#include <math.h>
#include <limits.h>
#include "total.h"
#include "template.h"
#include "init.h"
#include "symbol.h"
#include "dt.h"

static tree dt2tree(dt_t * dt);

#if D_GCC_VER < 40
// struct rtx was modified for c++; this macro from rtl.h needs to
// be modified accordingly.
#undef XEXP
#define XEXP(RTX, N)	(RTL_CHECK2 (RTX, N, 'e', 'u').rt_rtx)
#endif

IRState gen; // Top-level IRState
bool IRState::inGenObjFile;
ModuleInfo IRState::moduleInfo;
Array IRState::stmtExprList;
TemplateEmission IRState::emitTemplates;
bool IRState::warnSignCompare = false;
bool IRState::originalOmitFramePointer;
tree IRState::staticCtorList = NULL_TREE;

bool
d_gcc_force_templates()
{
    return IRState::emitTemplates == TEprivate ||
	IRState::emitTemplates == TEall;
}

void
IRState::doLineNote(const Loc & loc)
{
    if (loc.filename) {
	setLoc(loc);
#if D_GCC_VER >= 34
	location_t gcc_loc = { loc.filename, loc.linnum };
	emit_line_note (gcc_loc);
#else
	emit_line_note ((const char *)loc.filename, loc.linnum);
#endif
    } else {
	// do nothing
    }
}

void
IRState::setLoc(const Loc & loc)
{
    if (loc.filename) {
#if D_GCC_VER >= 34
	location_t gcc_loc = { loc.filename, loc.linnum };
	input_location = gcc_loc;
#else
	input_filename = loc.filename;
	lineno = loc.linnum;
#endif
    } else {
	// do nothing
    }
}

void IRState::setDeclLoc(tree t, const Loc & loc)
{
    // DWARF2 will often crash if the DECL_SOURCE_FILE is not set.  It's
    // easier the error here.
    assert(loc.filename);
    DECL_SOURCE_FILE (t) = loc.filename;
    DECL_SOURCE_LINE (t) = loc.linnum;
}

void
IRState::setDeclLoc(tree t, Dsymbol * decl)
{
    while (decl) {
	if (decl->loc.filename) {
	    setDeclLoc(t, decl->loc);
	    return;
	}
	decl = decl->toParent();
    }

    // fallback; backend sometimes crashes if not set
    
    Loc l;

    Module * m = decl->getModule();
    if (m && m->srcfile && m->srcfile->name) {
	l.filename = m->srcfile->name->str;
    } else {
	l.filename = "<no_file>"; // Emptry string can mess up debug info
    }    
    l.linnum = 1;
    setDeclLoc(t, l); 
}

void
IRState::setupSymbolStorage(Dsymbol * dsym, tree decl_tree, bool force_static_public)
{
    Declaration * real_decl = dsym->isDeclaration();
    FuncDeclaration * func_decl = real_decl ? real_decl->isFuncDeclaration() : 0;
    
    if (force_static_public ||
	( TREE_CODE( decl_tree ) == VAR_DECL && (real_decl && real_decl->isDataseg()) ) ||
	( TREE_CODE( decl_tree ) == FUNCTION_DECL ) ) {

	bool is_template = false;
	Dsymbol * sym = dsym->toParent();
	Module * ti_obj_file_mod;

	while (sym) {
	    TemplateInstance * ti = sym->isTemplateInstance();
	    if (ti) {
		ti_obj_file_mod = ti->objFileModule;
		is_template = true;
		break;
	    }
	    sym = sym->toParent();
	}

	bool is_static;

	if (is_template) {
	    D_DECL_ONE_ONLY( decl_tree ) = 1;
	    D_DECL_IS_TEMPLATE( decl_tree ) = 1;
	    is_static = ti_obj_file_mod == getCurrentModule() && emitTemplates != TEnone;
	} else {
	    is_static = dsym->getModule() == getCurrentModule();
	}
	/*
	if (func_decl)
	    fprintf(stderr, "%s: is_template = %d is_static = %d te = %d m = %s cur = %s\n",
		func_decl->toPrettyChars(), is_template, is_static, emitTemplates,
		dsym->getModule() ? dsym->getModule()->toChars() : "ZOMGNONE?",
		getCurrentModule()->toChars());
	*/
	
	if ( is_static ) {
	    DECL_EXTERNAL( decl_tree ) = 0;
	    TREE_STATIC( decl_tree ) = 1; // %% don't set until there is a body?
	} else {
	    DECL_EXTERNAL( decl_tree ) = 1;
	    TREE_STATIC( decl_tree ) = 0;
	}

	if (TREE_CODE( decl_tree ) == VAR_DECL &&
	    (real_decl && (real_decl->storage_class & STCextern)) )
	    is_static = false;

	if ( ( real_decl ) ) {
	    switch (real_decl->prot()) {
	    case PROTexport:
	    case PROTpublic:
	    case PROTpackage:
	    case PROTprotected:
		// %% set for specials like init,vtbl ? -- otherwise,special case
		// for reverse the default
		TREE_PUBLIC( decl_tree ) = 1;
		break;
	    default:
		if ( (func_decl && (func_decl->isMain() || func_decl->isWinMain() ||
			  func_decl->isDllMain())) ||
		    (real_decl->isMember() && ! real_decl->isThis())) { // %% check this -- static members/
		    // DMD seems to ignore private in this case...
		    TREE_PUBLIC( decl_tree ) = 1;
		} else {
		    // From DMD:
		    /* private statics should still get a global symbol, in case
		     * another module inlines a function that references it.
		     */

		    // TREE_PUBLIC( decl_tree ) = 0;
		    TREE_PUBLIC( decl_tree ) = 1;
		}
	    }
	}

	// %% Need to do this now, or Mach-O will have stub issues ?
	if (is_template && is_static && emitTemplates == TEprivate) {
	    TREE_PRIVATE(decl_tree) = 1;
	    TREE_PUBLIC(decl_tree) = 0;
	}

	if ( func_decl && func_decl->isNested() )
	    TREE_PUBLIC( decl_tree ) = 0;
    } else {
	TREE_STATIC( decl_tree ) = 0;
	DECL_EXTERNAL( decl_tree ) = 0;
	TREE_PUBLIC( decl_tree ) = 0;
    }
}
void
IRState::setupStaticStorage(Dsymbol * dsym, tree decl_tree)
{
    setupSymbolStorage(dsym, decl_tree, true);
    TREE_PUBLIC( decl_tree ) = 1;
}
void
IRState::prepareSymbolOutput(Symbol * s)
{
    tree decl_tree = s->Stree;
    /* make_decl_one_only checks variables for DECL_INITIAL.
       If it doesn't have one, it makes in a common variable.
       This is a problem because we don't have DECL_INITIAL
       until the variable is emitted.  Solution is to call this
       routine after DECL_INITIAL is set and before
       rest_of_decl_compilation. */
       
    if (D_DECL_ONE_ONLY(decl_tree)) {
	if (SUPPORTS_ONE_ONLY  // Must check, otherwise backend will abort

	    /* This check is only to make -femit-templates=private work on
	       targets that it is not actually needed. */
	    && ! (D_DECL_IS_TEMPLATE( decl_tree ) && emitTemplates == TEprivate)
	    ) {
	    // this sets TREE_PUBLIC
	    make_decl_one_only( decl_tree );
#ifdef MAKE_DECL_COALESCED
	    // for Apple gcc...
	} else {
	    // warn if coalescing disabled?
	    MAKE_DECL_COALESCED( decl_tree );
#else
	} else if ( ! D_DECL_IS_TEMPLATE( decl_tree ) || emitTemplates == TEprivate ) {
	    // Do this trick unconditionally for typeinfo.  For templates, check
	    // the emitTemplates setting.
	    TREE_PRIVATE(decl_tree) = 1;
	    TREE_PUBLIC(decl_tree) = 0;
	} else {
	    static bool warned = false;
	    if (! warned) {
		::warning("system does not support one-only linkage");
		warned = true;
	    }
#endif
	}
    }

    if (TREE_CODE(decl_tree) == FUNCTION_DECL &&
	DECL_STATIC_CONSTRUCTOR( decl_tree )) {
	    
	if (targetm.have_ctors_dtors) {
	    (* targetm.asm_out.constructor) (XEXP (DECL_RTL (decl_tree), 0),
		DEFAULT_INIT_PRIORITY);
	} else {
	    // change to list maker...
	    staticCtorList = tree_cons(NULL_TREE, decl_tree, staticCtorList);
	    // %%TODO: fallback if ! targetm.have_ctors_dtors
	}
    }
}

void
IRState::outputStaticSymbol(tree t) {
    d_add_global_function(t);

    // %%TODO: flags

    /* D allows zero-length declarations.  Such a declaration ends up with
       DECL_SIZE(t) == NULL_TREE which is what the backend function
       assembler_variable checks.  This could change in later versions...
       
       Maybe all of these variables should be aliased to one symbol... */
    
    // %%TODO: could move this to lang_hooks.decls.prepare_assemble_variable to
    // make this check less precarious. -- or finish_incomplete_decl (from
    // wrapup_global_declarations
    if ( DECL_SIZE(t) == 0 ) {
	DECL_SIZE(t) = bitsize_int(0);
	DECL_SIZE_UNIT(t) = size_int(0);
    } // Otherwise, if DECL_SIZE == 0, just let it fail...
    
    rodc(t, 1);
}


bool
IRState::shouldEmit(Dsymbol * d_sym)
{
    if (emitTemplates != TEnone)
	return true;

    Dsymbol * sym = d_sym->toParent();
    while (sym) {
	if (sym->isTemplateInstance()) {
	    return false;
	    break;
	}
	sym = sym->toParent();
    }
    return true;
}

bool
IRState::shouldEmit(Symbol * sym)
{
    return (emitTemplates != TEnone) || ! D_DECL_IS_TEMPLATE(sym->Stree);
}


void
IRState::emitLocalVar(VarDeclaration * v, bool uninit_okay)
{
    if (v->isDataseg() || (v->storage_class & STCfield))
	return;

    tree var_decl = v->toSymbol()->Stree;
    tree init_exp = NULL_TREE;
    tree cleanup  = NULL_TREE;
    FuncDeclaration * f = getCurrentFunction();

    assert( ! TREE_STATIC(var_decl ));

    // If this were enabled, it would cause each local to come into scope on the line it
    // is declared, but it is causing crashes... looks like expand_end_bindings is called too many times?
    // startBindings();

    DECL_CONTEXT( var_decl ) = getLocalContext();
    pushdecl(var_decl);
    
    // If not _argptr and there is not already an initializer created by
    // the glue code...
    if ( f->v_argptr == v ) {
	tree parm;

	// Get the last parameter
	for (parm = DECL_ARGUMENTS( f->toSymbol()->Stree );
	     TREE_CHAIN(parm) != NULL_TREE; parm = TREE_CHAIN(parm)) {
	    // nothing
	}

	init_exp = buildCall(void_type_node,
	    addressOf( built_in_decls[BUILT_IN_VA_START] ),
	    tree_cons(NULL_TREE, addressOf(var_decl),
		tree_cons( NULL_TREE, parm, NULL_TREE)));
	cleanup = buildCall(void_type_node,
	    addressOf( built_in_decls[BUILT_IN_VA_END] ),
	    tree_cons(NULL_TREE, addressOf(var_decl), NULL_TREE));
    } else if ( ! DECL_INITIAL( var_decl ) ) {
	if (v->init) {
	    if (! v->init->isVoidInitializer()) {
		ExpInitializer * exp_init = v->init->isExpInitializer();
		assert(exp_init);

		Expression * ie = exp_init->toExpression();
		do {
		    if (ie->op == TOKassign) {
			AssignExp * a_exp = (AssignExp *) ie;
			if (a_exp->e1->op == TOKvar &&
			    ((VarExp *) a_exp->e1)->var == v) {

			    DECL_INITIAL( var_decl ) = convertForAssignment(a_exp->e2, v->type);
			    break;
			}
		    }

		    init_exp = ie->toElem(this);
		} while (0);
	    } else
		uninit_okay = true;
	}
    }

    // %% Do we want to do this for special variables?
    if (inVolatile() && DECL_INITIAL( var_decl ) != NULL_TREE) {
	TREE_THIS_VOLATILE( DECL_INITIAL( var_decl )) = 1;
    }

    expand_decl(var_decl);
    // %%new: just forget this for now
    if (cleanup)
	expand_decl_cleanup(var_decl, cleanup);
	    
    // special case for _argptr
    if ( init_exp ) {
	gen.doLineNote(v->loc);
	expand_expr_stmt_value(init_exp, 0, 1);
    } else if ( DECL_INITIAL( var_decl )) {
	gen.doLineNote(v->loc);
	expand_decl_init(var_decl); // %%new: merge with above, irs->expandDecl
    } else if ( ! uninit_okay ) {
	// %% location
	warning("uninitialized variable '%s'", v->ident ? v->ident->string : "(no name)");
    }
}

tree
IRState::localVar(tree t_type) {
    tree t_decl = build_decl(VAR_DECL, NULL_TREE, t_type);
    DECL_CONTEXT( t_decl ) = getLocalContext();
    DECL_ARTIFICIAL( t_decl ) = 1;
    pushdecl(t_decl);
    return t_decl;
}


tree
IRState::declContext(Dsymbol * d_sym)
{
    Dsymbol * orig_sym = d_sym;
    AggregateDeclaration * agg_decl;
    
    while ( (d_sym = d_sym->toParent()) ) {
	if (d_sym->isFuncDeclaration()) {	    
	    // 3.3.x (others?) dwarf2out chokes without this check... (output_pubnames)
	    FuncDeclaration * f = orig_sym->isFuncDeclaration();
	    if (f && ! gen.functionNeedsChain(f))
		return 0;
	    return d_sym->toSymbol()->Stree;
	} else if ( (agg_decl = d_sym->isAggregateDeclaration()) ) {
	    ClassDeclaration * cd;
	    
	    tree ctx = agg_decl->type->toCtype();
	    if ( (cd = d_sym->isClassDeclaration()) ) {
		ctx = TREE_TYPE(ctx); // RECORD_TYPE instead of REFERENCE_TYPE

		/* Don't need this because backend uses decl_function_context(),
		which ignores the DECL_CONTEXT of a FUNC_DECL with
		DECL_VINDEX anyway. * /
		
		// Handle methods of a class nested in a function
		if (cd->isNested()) {
		    d_sym = cd->toParent();
		    if (d_sym->isFuncDeclaration()) {
			ctx = d_sym->toSymbol()->Stree;			
		    }
		}
		*/
	    }
	    return ctx;
	} else if ( d_sym->isModule() ) {
	    return d_sym->toSymbol()->ScontextDecl;
	}
    }
    return NULL_TREE;
}

void
IRState::expandDecl(tree t_decl)
{
#if D_GCC_VER < 40
    expand_decl(t_decl);
    if (DECL_INITIAL(t_decl))
	expand_decl_init(t_decl);
#else
    // nothing, pushdecl will add t_decl to a BIND_EXPR
    if (DECL_INITIAL(t_decl)) {
	// void_type_node%%?
	doExp(build2(MODIFY_EXPR, void_type_node, t_decl, DECL_INITIAL(t_decl)));
	DECL_INITIAL(t_decl) = NULL_TREE;
    }
#endif
}

void
IRState::addAggMethods(tree rec_type, AggregateDeclaration * agg)
{
    if (write_symbols != NO_DEBUG) {
	ListMaker methods;
	for (unsigned i = 0; i < agg->methods.dim; i++) {
	    FuncDeclaration * fd = (FuncDeclaration *) agg->methods.data[i];
	    
	    methods.chain(fd->toSymbol()->Stree);
	}
	TYPE_METHODS( rec_type ) = methods.head;
    }
}

void IRState::declareType(tree t, Type * d_type)
{
    // Note: not safe to call d_type->toCtype()
    Loc l;
    tree decl = build_decl(TYPE_DECL, get_identifier(d_type->toChars()), t);
    l.filename = "<internal>";
    l.linnum = 1;
    setDeclLoc(decl, l);
    // %% toplevel context

    declareType(t, decl);
}

void IRState::declareType(tree t, Dsymbol * d_sym)
{
    assert( ! POINTER_TYPE_P( t ) );

    const char * name = d_sym->ident ? d_sym->ident->string : "fix";
    tree decl = build_decl(TYPE_DECL, get_identifier(name), t);
    DECL_CONTEXT( decl ) = gen.declContext(d_sym);
    setDeclLoc(decl, d_sym);
    
    declareType(t, decl);
}


void
IRState::declareType(tree t, tree decl)
{
    assert( ! POINTER_TYPE_P( t ) );
    
    bool top_level =
	! DECL_CONTEXT(decl)
#if D_GCC_VER >= 34
	|| TREE_CODE(DECL_CONTEXT(decl))==TRANSLATION_UNIT_DECL
#endif
	;
    TYPE_CONTEXT( t ) = DECL_CONTEXT( decl );

    /* The following dance is needed to make typedefs work for DWARF 2.
       C and C++ do this different ways. The C version crashes for dbx (when
       member functions are involved) and I can't get the C++ way to work.
       So here's a third way. */
    if (1) {
	tree t_clone = build_type_copy( t );
	TYPE_MAIN_VARIANT(t_clone) = t_clone;
	TYPE_NEXT_VARIANT(t_clone) = NULL_TREE;
	TYPE_NEXT_VARIANT( t ) = NULL_TREE;

	tree clone_stub = build_decl(TYPE_DECL, NULL_TREE, t_clone);
	
	DECL_CONTEXT( clone_stub ) = DECL_CONTEXT( decl );
	// TREE_CHAIN( clone_stub ) = decl; // not needed?

	TYPE_STUB_DECL( t_clone ) = clone_stub;

	// rest_of_type_compilation(t_clone, 1); // not needed?
	DECL_ORIGINAL_TYPE( decl ) = t_clone;
    }

    TYPE_NAME( t ) = decl;
    TYPE_STUB_DECL( t ) = decl;

    rodc(decl, top_level);
    //pushdecl(decl);
}

tree
IRState::convertTo(Expression * exp, Type * target_type)
{
    return convertTo(exp->toElem( this ), exp->type, target_type);
}

tree
IRState::convertTo(tree exp, Type * exp_type, Type * target_type)
{
    tree result = 0;
    target_type = target_type->toBasetype();

    assert(exp_type);
    exp_type = exp_type->toBasetype();
    
    if (exp_type->equals(target_type))
	return exp;

    switch (exp_type->ty) {
    case Tdelegate:
	// %%TODO:NOP/VIEW_CONVERT
	if (target_type->ty == Tdelegate) {
	    return delegateVal( delegateMethodRef(exp), delegateObjectRef(exp),
		target_type);
	} else {
	    ::error("can't convert a delegate expression to %s", target_type->toChars());
	    return error_mark_node;
	}
	break;
    case Tclass:
	if (target_type->ty == Tclass) {
	    ClassDeclaration * target_class_decl = ((TypeClass *) target_type)->sym;
	    ClassDeclaration * obj_class_decl = ((TypeClass *) exp_type)->sym;
	    bool use_dynamic = false;

	    int offset;
	    if (target_class_decl->isBaseOf(obj_class_decl, & offset)) {
		// Casting up the inheritance tree: Don't do anything special.
		// Cast to an implemented interface: Handle at compile time.
		if (offset == OFFSET_RUNTIME) {
		    use_dynamic = true;
		} else if (offset) {
		    tree t = target_type->toCtype();
		    exp = maybeMakeTemp(exp);
		    return build(COND_EXPR, t,
			boolOp(NE_EXPR, exp, nullPointer()),
			build(PLUS_EXPR, t, exp, size_int(offset)),
			nullPointer());
		} else {
		    // d_convert will make a NOP cast
		    break;
		}
	    } else if ( target_class_decl == obj_class_decl ) {
		// d_convert will make a NOP cast
		break;
	    } else if ( ! obj_class_decl->isCOMclass() ) {
		use_dynamic = true;
	    }
	    
	    if (use_dynamic) {
		// Otherwise, do dynamic cast
		tree args[2] = { exp, addressOf( target_class_decl->toSymbol()->Stree ) }; // ##v (and why not just addresOf(target_class_decl)
		return libCall(obj_class_decl->isInterfaceDeclaration()
		    ? LIBCALL_INTERFACE_CAST : LIBCALL_DYNAMIC_CAST, 2, args);
	    } else {
		warning("cast to %s will fail", target_type->toChars());
		result = convert(target_type->toCtype(), d_null_pointer);
		if (TREE_SIDE_EFFECTS( exp )) { // make sure the expression is still evaluated if user expects it
		    result = compound(exp, result);
		}
		return result;
	    }
	} else {
	    // nothing; default
	}
	break;
    case Tsarray:
	{
	    // %% TODO: stuffing types ok?
	    tree pointer_value = addressOf( exp );
	    if (target_type->ty == Tpointer) {
		result = pointer_value;
		TREE_TYPE(result) = target_type->toCtype();
	    } else if (target_type->ty == Tarray) {
		TypeSArray * a_type = (TypeSArray*) exp_type;
		
		unsigned array_len = a_type->dim->toUInteger();
		unsigned sz_a = a_type->next->size();
		unsigned sz_b = target_type->next->size();

		// conversions to different sizes
		// Assumes tvoid->size() == 1
		// %% TODO: handle misalign like _d_arraycast_xxx ?
		if (a_type->next->isbit() != target_type->next->isbit()) {
		    if (a_type->next->isbit()) {
			array_len /= 8 * sz_b;
		    } else {
			array_len *= 8 * sz_a;
		    }
		} else if (sz_a != sz_b) {
		    array_len = array_len * sz_a / sz_b;
		}
		TREE_TYPE(pointer_value) = target_type->next->pointerTo()->toCtype();

		// Assumes casting to dynamic array of same type or void
		return darrayVal(target_type, array_len, pointer_value);
	    }
	    // %% else error?
	}
	break;
    case Tarray:
	if (target_type->ty == Tpointer) {
	    return convert(target_type->toCtype(), darrayPtrRef( exp ));
	} else if (target_type->ty == Tarray) {
	    // assume tvoid->size() == 1

	    Type * src_elem_type = exp_type->next->toBasetype();
	    Type * dst_elem_type = target_type->next->toBasetype();
	    unsigned sz_a = src_elem_type->size();
	    unsigned sz_b = dst_elem_type->size();
	    
	    if ((sz_a  != sz_b) ||
		 src_elem_type->isbit() != dst_elem_type->isbit()) {
		
		//TY exp_elem_ty = exp_type->next->toBasetype()->ty;
		if (! src_elem_type->isbit()) {
		    tree args[3] = {
			// assumes Type::tbit->size() == 1
			integerConstant(sz_b, Type::tuns32),
			integerConstant(sz_a * (dst_elem_type->isbit() ? 8 : 1),
			    Type::tuns32),
			exp
		    };
		    return libCall(LIBCALL_ARRAYCAST, 3, args, target_type->toCtype());
		} else {
		    tree args[2] = {
			integerConstant(sz_b, Type::tuns32),
			exp
		    };
		    return libCall(LIBCALL_ARRAYCAST_FROMBIT, 2, args, target_type->toCtype());
		}
	    } else {
		// %% VIEW_CONVERT_EXPR or NOP_EXPR ? (and the two cases below)
		// Convert to/from void[] or elements are the same size -- don't change length
		return build1(NOP_EXPR, target_type->toCtype(), exp);
	    }
	} else if (target_type->ty == Taarray) {
	    // DMD allows this.
	    return build1(NOP_EXPR, target_type->toCtype(), exp);
	}
	// else, default conversion, which should produce an error
	break;
    case Taarray:
	if (target_type->ty == Taarray || target_type->ty == Tarray) {
	    // DMD allows this.
	    return build1(NOP_EXPR, target_type->toCtype(), exp);
	}
	// else, default conversion, which should product an error
	break;
    default:
	if (exp_type->isreal() && target_type->isimaginary() ||
	    exp_type->isimaginary() && target_type->isreal()) {
	    // warn? handle in front end?
	    
	    result = build_real_from_int_cst( target_type->toCtype(), integer_zero_node );
	    if (TREE_SIDE_EFFECTS( exp ))
		result = compound(exp, result);
	    return result;
	} else if (exp_type->iscomplex()) {
	    Type * part_type;
	    // creal.re, .im implemented by cast to real or ireal
	    // Assumes target type is the same size as the original's components size
	    if (target_type->isreal()) {
		// maybe install lang_specific...
		switch (exp_type->ty) {
		case Tcomplex32: part_type = Type::tfloat32; break;
		case Tcomplex64: part_type = Type::tfloat64; break;
		case Tcomplex80: part_type = Type::tfloat80; break;
		default:
		    abort();
		}
		result = realPart(exp);
	    } else if (target_type->isimaginary()) {
		switch (exp_type->ty) {
		case Tcomplex32: part_type = Type::timaginary32; break;
		case Tcomplex64: part_type = Type::timaginary64; break;
		case Tcomplex80: part_type = Type::timaginary80; break;
		default:
		    abort();
		}
		result = imagPart(exp);
	    } else {
		// default conversion
		break;
	    }
	    result = convertTo(result, part_type, target_type);
	} else if (target_type->iscomplex()) {
	    tree c1, c2, t;
	    c1 = convert(TREE_TYPE(target_type->toCtype()), exp);
	    c2 = build_real_from_int_cst( TREE_TYPE(target_type->toCtype()), integer_zero_node);
	    
	    if (exp_type->isreal()) {
		// nothing
	    } else if (exp_type->isimaginary()) {
		t = c1;
		c1 = c2;
		c2 = t;
	    } else {
		// default conversion
		break;
	    }
	    result = build(COMPLEX_EXPR, target_type->toCtype(), c1, c2);
	} else {
	    assert( TREE_CODE( exp ) != STRING_CST );
	    // default conversion
	}
    }

    if (! result)
	result = d_convert_basic(target_type->toCtype(), exp);
    return result;
}

tree
IRState::convertForInitialization(Expression * exp, Declaration * target_decl)
{
    return doConvertForInit(exp, NULL_TREE, exp->type, target_decl);
}

tree
IRState::convertForInitialization(tree exp_tree, Type * exp_type, Declaration * target_decl)
{
    return doConvertForInit(0, exp_tree, exp_type, target_decl);
}

// Link convertForInitialization, but uses the information from Argument instead
// of a Declaration.
tree
IRState::convertForArgument(Expression * exp, Argument * arg)
{
    if ( isArgumentReferenceType(arg) ) {
	tree exp_tree = this->toElemLvalue(exp);
	// front-end already sometimes automatically takes the address
	// TODO: find out why... and can this be confused by a non-zero SymOff
	if (exp->op != TOKaddress && exp->op != TOKsymoff)
	    return addressOf( exp_tree );
	else
	    return exp_tree;
    } else {
	return convertForAssignment( exp->toElem(this), exp->type, arg->type );
    }
}

tree
IRState::doConvertForInit(Expression * exp, tree in_exp_tree, Type * in_exp_type, Declaration * target_decl)
{
    if ( isDeclarationReferenceType( target_decl ) ) {
	// %% not converting to exact reference type...
	return addressOf( exp ? exp->toElem(this) : in_exp_tree );
    } else {
	tree result;
	
	if ( exp ) {
	    result = convertForAssignment(exp, target_decl->type);
	} else {
	    result = convertForAssignment(in_exp_tree, in_exp_type, target_decl->type);
	}

	// We can't init static storage with a VAR_DECL (probably can't
	// do it with an auto variable, either -- but can just use
	// MODIFY_EXPR in that case).  Also applies to fields of
	// record in static storage
	
	// %%TODO: Are these all the cases?
	
	if (! target_decl->type->isscalar() &&
	    ( target_decl->isDataseg() ||
	      (target_decl->storage_class & STCfield)) ) {
	    result = stripVarDecl( result );
	}
	return result;
    }
}


// Apply semantics of assignment to a values of type <target_type> to <exp>
// (e.g., pointer = array -> pointer = & array[0])

// Expects base type to be passed
static Type *
final_sa_elem_type(Type * type)
{
    while (type->ty == Tsarray) {
	type = type->next->toBasetype();
    }
    return type;
}

tree
IRState::convertForAssignment(Expression * exp, Type * target_type)
{
    Type * exp_base_type = exp->type->toBasetype();
    Type * target_base_type = target_type->toBasetype();

    // Assuming this only has to handle converting a non Tsarray type to
    // arbitrarily dimensioned Tsarrays.
    if (target_base_type->ty == Tsarray &&
	final_sa_elem_type(target_base_type)->equals(exp_base_type)) { // %% what about implicit converions...?

	TypeSArray * sa_type = (TypeSArray *) target_base_type;
	uinteger_t count = sa_type->dim->toUInteger();
	
	tree ctor = make_node( CONSTRUCTOR );
	TREE_TYPE( ctor ) = target_type->toCtype();
	if (count) {
	    CONSTRUCTOR_ELTS( ctor ) = tree_cons( build(RANGE_EXPR, Type::tsize_t->toCtype(),
						      integer_zero_node, integerConstant( count - 1 )),
		stripVarDecl(convertForAssignment(exp, sa_type->next)), NULL_TREE );
	}
	TREE_READONLY( ctor ) = 1;
	TREE_CONSTANT( ctor ) = 1;
	return ctor;
    } else if (! target_type->isscalar() && exp_base_type->isintegral()) {
	// D Front end uses IntegerExp(0) to mean zero-init a structure
	// This could go in convert for assignment, but we only see this for
	// internal init code -- this also includes default init for _d_newarrayi...
	
	if (exp->toInteger() == 0) {
	    tree empty = make_node( CONSTRUCTOR );
	    TREE_TYPE( empty ) = target_type->toCtype();
	    CONSTRUCTOR_ELTS( empty ) = NULL_TREE; // %% will this zero-init?
	    TREE_CONSTANT( empty ) = 1;
	    // static?
	    return empty;
	    // %%TODO: Use a code (lang_specific in decl or flag) to memset instead?
	} else {
	    abort();
	}
    }
    
    tree exp_tree = exp->toElem(this);
    return convertForAssignment(exp_tree, exp->type, target_type);
}

tree
IRState::convertForAssignment(tree expr, Type * expr_type, Type * target_type)
{
    return convertTo(expr, expr_type, target_type);
}

// could be like C c-typeck.c:default_conversion
// todo:

/* Perform default promotions for C data used in expressions.
   Arrays and functions are converted to pointers;
   enumeral types or short or char, to int.
   In addition, manifest constants symbols are replaced by their values.  */

// what about float->double?

tree
IRState::convertForCondition(tree exp_tree, Type * exp_type) {
    tree result = exp_tree;
    switch (exp_type->toBasetype()->ty) {
    case Taarray:
    case Tarray:
	result = darrayLenRef(result);
	break;
    case Tdelegate:
	// DMD checks (function || object), but what good
	// is if if there is a null function pointer?
	{
	    tree a, b;
	    if ( D_IS_METHOD_CALL_EXPR(result) ) {
		extractMethodCallExpr(result, a, b);
	    } else {
		tree tmp = maybeMakeTemp(result);
		a = delegateObjectRef(tmp);
		b = delegateMethodRef(tmp);
	    }
	    // not worth using  or TRUTH_ORIF...
	    // %%TODO: Is this okay for all targets?
	    result = build(BIT_IOR_EXPR, TREE_TYPE(a), a, b);
	}
	break;
    default:
	break;
    }
    // see expr.c:do_jump for the types of trees that can be passed to expand_start_cond
    //TREE_USED( <result> ) = 1; // %% maybe not.. expr optimized away because constant expression?
    return d_truthvalue_conversion( result );
}



tree
IRState::convertForVarArg(Expression * exp)
{
    // The conversions are already done by the front end
    // This just handles stuff we build.
    
    if (exp->type) {
	exp = exp->integralPromotions();
	// %% figure out why I needed this...
	if (exp->type->toBasetype()->ty == Tsarray)
	    exp = exp->castTo( exp->type->toBasetype()->next->arrayOf() );
    }
    return exp->toElem( this );
}

/* Convert to void[] without changing the length */
tree
IRState::rawArray(Expression * exp)
{
    TY ty = exp->type->toBasetype()->ty;
    tree val;
    if (ty == Tsarray) {
	val = convertTo(exp, exp->type->toBasetype()->next->arrayOf());
    } else if (ty == Taarray || ty == Tarray) {
	val = exp->toElem(this);
    } else {
	abort();
	return 0;
    }
    return build1(NOP_EXPR, Type::tvoid->arrayOf()->toCtype(), val);
}


tree
IRState::stripVarDecl(tree value) {
    if ( TREE_CODE( value ) != VAR_DECL ) {
	return value;
    } else if ( DECL_INITIAL(value) ) {
	return DECL_INITIAL(value);
    } else {
	Type * d_type = getDType( TREE_TYPE( value ));
	if (d_type) {
	    d_type = d_type->toBasetype();
	    switch (d_type->ty) {
	    case Tstruct:
		{
		    //tree t = gen.aggregateInitializer(((TypeStruct *) d_type)->sym, NULL, NULL, NULL_TREE);
		    // %% maker sure this is doing the right thing...
		    // %% better do get rid of it..

		    // need to VIEW_CONVERT?
		    dt_t * dt = NULL;
		    d_type->toDt(& dt);
		    tree t = dt2tree( dt );
		    TREE_CONSTANT(t) = 1;
		    return t;
		}
		break;
	    default:
		// error below
		break;
	    }
	}
	abort();
    }
}

tree
IRState::pointerIntSum(tree ptr_node, tree idx_exp)
{
    tree result_type_node = TREE_TYPE( ptr_node );
    tree intop = idx_exp;
    tree size_exp;

    // %% TODO: real-not-long-double issues...

    // %% test for void case ...
    size_exp = size_in_bytes( TREE_TYPE( result_type_node ) ); // array element size
    if (integer_zerop(size_exp))
	size_exp = integer_one_node;

    if (TYPE_PRECISION (TREE_TYPE (intop)) != TYPE_PRECISION (sizetype)
	|| TREE_UNSIGNED (TREE_TYPE (intop)) != TREE_UNSIGNED (sizetype))
	intop = convert (d_type_for_size (TYPE_PRECISION (sizetype), 
			     TREE_UNSIGNED (sizetype)), intop);
    
    intop = convert (result_type_node,
	build/*_binary_op*/ (MULT_EXPR, TREE_TYPE(size_exp), intop,  // the type here may be wrong %%
	    convert (TREE_TYPE (intop), size_exp)));
    intop = fold(intop);

    if ( integer_zerop(intop) )
	return ptr_node;
    else
	return build(PLUS_EXPR, result_type_node, ptr_node, intop);
}

// Doesn't do some of the omptimizations in ::makeArrayElemRef
tree
IRState::checkedIndex(Loc loc, tree index, tree upper_bound, bool inclusive)
{
    if (global.params.useArrayBounds) {
	return build(COND_EXPR, TREE_TYPE(index),
	    boundsCond(index, upper_bound, inclusive),
	    index,
	    assertCall(loc, LIBCALL_ARRAY_BOUNDS));
    } else {
	return index;
    }
}


// index must be wrapped in a SAVE_EXPR to prevent multiple evaluation...
tree
IRState::boundsCond(tree index, tree upper_bound, bool inclusive)
{
    tree bound_check;

    bound_check = build(inclusive ? LE_EXPR : LT_EXPR, boolean_type_node,
	convert( d_unsigned_type(TREE_TYPE(index)), index ),
	upper_bound);
    
    if (! TREE_UNSIGNED( TREE_TYPE( index ))) {
	bound_check = build(TRUTH_ANDIF_EXPR, boolean_type_node, bound_check,
	    // %% conversions
	    build(GE_EXPR, boolean_type_node, index, integer_zero_node));
    }

    return bound_check;
}

tree
IRState::bitAlignCheck(Loc loc, tree index)
{
    tree cond;
    
    index = maybeMakeTemp(index);

    cond = build(BIT_AND_EXPR, TREE_TYPE(index),
	index, integerConstant(7, TREE_TYPE(index)));
    cond = build(EQ_EXPR, boolean_type_node, cond, integer_zero_node);

    return build(COND_EXPR, TREE_TYPE(index), cond, index,
	assertCall(loc, LIBCALL_ARRAY_BOUNDS));
}


tree
IRState::assertCall(Loc loc, LibCall libcall)
{
    tree args[2] = { darrayString(loc.filename ? loc.filename : ""),
		     integerConstant(loc.linnum, Type::tuns32) };
    return libCall(libcall, 2, args);
}

tree
IRState::floatConstant( real_t value, TypeBasic * target_type )
{
    REAL_VALUE_TYPE converted_val;
    
    tree type_node = target_type->toCtype();
    real_convert(& converted_val, TYPE_MODE(type_node), & value.rv());
    
    return build_real(type_node, converted_val);
}

void
IRState::grokSlice(Expression * slice_exp, tree & array_out, tree & lwr_out, tree & upr_out, int flags)
{
    Expression * array_exp, * lwr_exp, * upr_exp;
    extractSliceExp(slice_exp, array_exp, lwr_exp, upr_exp);

    array_out = array_exp->toElem(this);

    if (lwr_exp) {
	lwr_out = lwr_exp->toElem(this);
	if (flags & GS_BoundsCheck && global.params.useArrayBounds) {
	    // and not ((array-is-static && lwr-is-constant) or ( lwr-is-const-zero ))
	    abort(); // todo
	}
    } else if (flags & GS_CreateBounds)
	lwr_out = integerConstant(0, Type::tindex);
    else
	lwr_out = NULL_TREE;
    if (upr_exp) {
	upr_out = upr_exp->toElem(this);
	if (flags & GS_BoundsCheck && global.params.useArrayBounds) {
	    abort(); // todo
	}
    } else if (flags & GS_CreateBounds) {
	array_out = maybeMakeTemp(array_out);
	upr_out = arrayLength(array_out, array_exp->type);
    }

    Type * array_base_type = array_exp->type->toBasetype();

    switch (flags & GS_ProcessMask) {
    case GS_ExtractPointer:
	switch (array_base_type->ty) {
	case Tsarray:
	case Tarray:
	    array_out = convertTo(array_out, array_base_type, array_base_type->next->pointerTo());
	    break;
	case Tpointer:
	    // nothing;
	    break;
	default:
	    slice_exp->error("Unexpected type for slice: %s", array_base_type->toChars());
	    lwr_out = upr_out = array_out = error_mark_node;
	    return;
	}
	break;
    case 0:
	break;
    default:
	abort();
    }
}

tree
IRState::call(FuncDeclaration * func_decl, Array * args)
{
    Loc loc;
    Expression * exp = new VarExp(loc, func_decl);
    exp->type = func_decl->type;
    return call(exp, /*(TypeFunction *) func_decl->type, */ args);
}

tree
IRState::libCall(LibCall lib_call, unsigned n_args, tree *args, tree force_result_type)
{
    tree result;
    tree callee = functionPointer( getLibCallDecl( lib_call ));
    // for force_result_type, assumes caller knows what it is doing %%
    tree result_type = force_result_type != NULL_TREE ?
	force_result_type : TREE_TYPE(TREE_TYPE(TREE_OPERAND(callee, 0)));
    tree arg_list = NULL_TREE;
    for (int i = n_args - 1; i >= 0; i--) {
	arg_list = tree_cons(NULL_TREE, args[i], arg_list);
    }

    result = buildCall(result_type, callee, arg_list);
    return result;
}

static inline bool
function_type_p(tree t)
{
    return TREE_CODE(t) == FUNCTION_TYPE || TREE_CODE(t) == METHOD_TYPE;
}

tree
IRState::call(tree callable, tree object, Array * arguments)
{
    tree func_type_node = TREE_TYPE( callable );
    
    tree actual_callee;
    TypeFunction * func_type = 0;
    
    if ( POINTER_TYPE_P( func_type_node ) ) {
	func_type_node = TREE_TYPE( func_type_node );
	actual_callee = callable;
    } else {
	actual_callee = addressOf( callable );
    }

    assert( function_type_p( func_type_node ) );
    assert( TYPE_LANG_SPECIFIC( func_type_node ) != 0 );
    func_type = (TypeFunction *) getDType(func_type_node)->toBasetype();
    assert( func_type->ty == Tfunction );

    bool is_d_vararg = func_type->varargs == 1 && func_type->linkage == LINKd;

    ListMaker actual_arg_list;
    tree result;

    if ( TREE_CODE( func_type_node ) == FUNCTION_TYPE ) {
	if ( object != NULL_TREE

	    && ! TYPE_LANG_SPECIFIC( func_type_node )->is_nested_function // %% check

	     ) {
	    
	    tree method_type = build_method_type( TREE_TYPE( object ), func_type_node );
	    TYPE_ATTRIBUTES( method_type ) = TYPE_ATTRIBUTES( func_type_node );
	    func_type_node = method_type;
	}
    } else /* METHOD_TYPE */ {
	if ( ! object ) {
	    // Front-end apparently doesn't check this.
	    if (TREE_CODE(callable) == FUNCTION_DECL ) {
		error("need 'this' to access member %s", IDENTIFIER_POINTER( DECL_NAME( callable )));
		return error_mark_node;
	    } else {
		// Probably an internal error
		
		assert(object != NULL_TREE);
	    }
	}
    }

    Array * formal_args = func_type->arguments; // can be NULL for genCfunc decls
    unsigned n_formal_args = formal_args ? formal_args->dim : 0;
    int n_actual_args = arguments ? arguments->dim : 0;
    int fa_adj = is_d_vararg ? -1 : 0;

    // assumes arguments->dim <= formal_args->dim if (! this->varargs)
    for (int i = - 1; i < n_actual_args; i++) {
	tree actual_arg_tree;

	if (i == -1) {
	    // If this is a nested function being called as a delegate,
	    // object should not be NULL_TREE and will be the dummy argument.
	    // If calling a nested function directly, pass zero for the
	    // dummy argument.
	    if (object != NULL_TREE) {
		actual_arg_tree = object;
	    } else if ( TYPE_LANG_SPECIFIC( func_type_node )->is_nested_function ) {
#if D_NO_TRAMPOLINES
		tree t = callable;
		if (POINTER_TYPE_P(TREE_TYPE(callable)))
		    t = TREE_OPERAND(callable, 0);
		actual_arg_tree = getStaticChain(t);
#else	
		actual_arg_tree = integer_zero_node;
#endif
		//fprintf(stderr, "does this happen?\n");
	    } else {
		continue;
	    }
	} else {
	    Expression * actual_arg_exp = (Expression *) arguments->data[i];
	    if (i == 0 && is_d_vararg) {
		// The hidden _arguments parameter
		actual_arg_tree = actual_arg_exp->toElem(this);
	    } else if ((unsigned)(i+fa_adj) < n_formal_args) {
		// Actual arguments for declared formal arguments
		Argument * formal_arg = (Argument *) formal_args->data[i + fa_adj];

		actual_arg_tree = convertForArgument(actual_arg_exp, formal_arg);
	    } else if (func_type->linkage != LINKd) {
		// Vararg arguments
		actual_arg_tree = convertForVarArg(actual_arg_exp);
	    } else {
		actual_arg_tree = actual_arg_exp->toElem(this);
	    }
	}

	//TREE_USED( actual_arg_tree ) = 1; // needed ?
	
	actual_arg_list.cons( actual_arg_tree );
    }

    result = buildCall(TREE_TYPE(func_type_node), actual_callee, actual_arg_list.head);
    return maybeExpandSpecialCall(result);
}

tree
IRState::call(Expression * expr, /*TypeFunction * func_type, */ Array * arguments)
{
    tree callee = expr->toElem(this);
    tree object = NULL_TREE;

    if ( D_IS_METHOD_CALL_EXPR( callee ) ) {
	// This could be a delegate expression (TY == Tdelegate), but not
	// actually a delegate variable.
	
	extractMethodCallExpr(callee, callee, object);
    } else if (expr->type->toBasetype()->ty == Tdelegate) {
	object = delegateObjectRef(callee);
	callee = delegateMethodRef(callee);
    } else if (expr->op == TOKvar) {
	FuncDeclaration * func = ((VarExp *) expr)->var->isFuncDeclaration();
	if (func && func->isNested()) {
#if D_NO_TRAMPOLINES
	    object = getStaticChain(func);
#else
	    // Pass fake argument for nested functions
	    object = d_null_pointer;
#endif
	}
    }
    return call(callee, object, arguments);
}

tree
IRState::libCall(LibCall lib_call, unsigned n_args, Expression ** args, tree result_type)
{
    Array a;
    a.setDim(n_args);
    memmove(a.data, args, n_args * sizeof(void *));
    return libCall(lib_call, & a, result_type);
}

static const char * libcall_ids[LIBCALL_count] =
    { "_d_assert", "_d_array_bounds", "_d_switch_error", "_D9invariant12_d_invariantFC6ObjectZv",
      "_d_newclass", "_d_new","_d_newarrayi", "_d_newbitarray",
      "_d_delclass", "_d_delarray", "_d_delmemory", "_d_callfinalizer",
      "_d_arraysetlength", "_d_arraysetlengthb",
      "_d_dynamic_cast", "_d_interface_cast",
      "_adEq", "_adEqBit", "_adCmp", "_adCmpChar", "_adCmpBit",
      "_aaIn", "_aaGet", "_aaGetRvalue", "_aaDel",
      "_d_arraycast", "_d_arraycast_frombit",
      "_d_arraycopy", "_d_arraycopybit",
      "_d_arraycat", "_d_arraycatb", "_d_arraycatn",
      "_d_arrayappend", "_d_arrayappendb",
      "_d_arrayappendc", "_d_arrayappendcb",
      "_d_arraysetbit", "_d_arraysetbit2",
      "_d_monitorenter", "_d_monitorexit",
      "_d_criticalenter", "_d_criticalexit",
      "_d_throw",
      "_d_switch_string", "_d_switch_ustring", "_d_switch_dstring"
    };

static FuncDeclaration * libcall_decls[LIBCALL_count];

void
IRState::replaceLibCallDecl(FuncDeclaration * d_decl)
{
    if ( ! d_decl->ident )
	return;
    for (unsigned i = 0; i < LIBCALL_count; i++) {
	if ( strcmp(d_decl->ident->string, libcall_ids[i]) == 0 ) {
	    // %% warn if libcall already set?
	    // Only do this for the libcalls where it's a problem, otherwise
	    // it causes other problems...
	    switch ((LibCall) i) {
		// case LIBCALL_GNU_BITARRAYSLICEP:
	    case LIBCALL_ARRAYCOPY: // this could be solved by turning copy of char into memcpy
	    case LIBCALL_ARRAYCAST:
		// replace the function declaration
		break;
#if D_GCC_VER <= 33
		// may be safe for GCC 3.4, haven't tried it
	    case LIBCALL_ADCMPCHAR:
#endif
	    default:
		// don't replace
		return;
	    }
	    libcall_decls[i] = d_decl;
	    break;
	}
    }
}


FuncDeclaration *
IRState::getLibCallDecl(LibCall lib_call)
{
    FuncDeclaration * decl = libcall_decls[lib_call];
    Array arg_types;
    Type * t = NULL;
    bool varargs = false;

    if (! decl) {
	Type * return_type = Type::tvoid;
	
	switch (lib_call) {
	case LIBCALL_ASSERT:
	case LIBCALL_ARRAY_BOUNDS:
	case LIBCALL_SWITCH_ERROR:
	    // need to spec chararray/string because internal code passes string constants
	    arg_types.reserve(2);
	    arg_types.push( Type::tchar->arrayOf() );
	    arg_types.push( Type::tuns32 );
	    break;
	case LIBCALL_NEWCLASS:
	    arg_types.push( ClassDeclaration::classinfo->type );
	    return_type = getObjectType();
	    break;
	case LIBCALL_NEW:
	    arg_types.push( Type::tuns32 );
	    arg_types.push( Type::tuns32 );
	    return_type = Type::tvoid->arrayOf();
	    break;
	case LIBCALL_NEWARRAYI:
	    arg_types.push( Type::tuns32 );
	    arg_types.push( Type::tuns32 );
	    varargs = true;
	    return_type = Type::tvoid->arrayOf();
	    break;
	case LIBCALL_NEWBITARRAY:
	    arg_types.push( Type::tuns32 );
	    arg_types.push( Type::tbit );
	    return_type = Type::tbit->arrayOf(); 
	    break;
	case LIBCALL_DELCLASS:
	    arg_types.push(getObjectType()->pointerTo());
	    break;
	case LIBCALL_DELARRAY:
	    arg_types.push(Type::tvoid->arrayOf()->pointerTo());
	    break;
	case LIBCALL_DELMEMORY:
	    arg_types.push(Type::tvoid->pointerTo()->pointerTo());
	    break;
	case LIBCALL_CALLFINALIZER:
	    arg_types.push(Type::tvoid->pointerTo()); // declared void*; could be Object....
	    break;
	case LIBCALL_ARRAYSETLENGTH:
	case LIBCALL_ARRAYSETLENGTH_B:
	    arg_types.push( Type::tuns32 );
	    if (lib_call == LIBCALL_ARRAYSETLENGTH)
		arg_types.push( Type::tuns32 );
	    arg_types.push( Type::tvoid->arrayOf()->pointerTo() );
	    return_type = Type::tvoid->arrayOf(); 
	    break;
	case LIBCALL_DYNAMIC_CAST:
	case LIBCALL_INTERFACE_CAST:
	    arg_types.push( getObjectType() );
	    arg_types.push( ClassDeclaration::classinfo->type );
	    return_type = getObjectType();
	    break;
	case LIBCALL_ADEQ:
	case LIBCALL_ADCMP:
	    arg_types.reserve(3);
	    arg_types.push(Type::tvoid->arrayOf());
	    arg_types.push(Type::tvoid->arrayOf());
	    arg_types.push(Type::typeinfo->type);
	    return_type = Type::tint32;
	    break;
	case LIBCALL_ADCMPCHAR:
	    arg_types.reserve(2);
	    arg_types.push(Type::tchar->arrayOf());
	    arg_types.push(Type::tchar->arrayOf());
	    return_type = Type::tint32;
	    break;
	case LIBCALL_ADEQBIT:
	case LIBCALL_ADCMPBIT:
	    arg_types.reserve(2);
	    arg_types.push(Type::tbit->arrayOf());
	    arg_types.push(Type::tbit->arrayOf());
	    return_type = Type::tint32;
	    break;
	case LIBCALL_AAIN:
	case LIBCALL_AAGET:
	case LIBCALL_AAGETRVALUE:
	case LIBCALL_AADEL:
	    {
		Type * aa_type = Type::tvoid->pointerTo()->arrayOf(); // associated arrays are dynamic arrays of pointers
		if (lib_call == LIBCALL_AAGET)
		    aa_type = aa_type->pointerTo();
		
		arg_types.reserve(3);
		arg_types.push(aa_type);
		arg_types.push(Type::typeinfo->type); // typeinfo reference
		if ( lib_call == LIBCALL_AAGET || lib_call == LIBCALL_AAGETRVALUE)
		    arg_types.push(Type::tint32); // %%
		
		varargs = true;
		
		switch (lib_call) {
		case LIBCALL_AAIN:
		case LIBCALL_AAGET:
		case LIBCALL_AAGETRVALUE:
		    return_type = Type::tvoid->pointerTo();
		    break;
		case LIBCALL_AADEL:
		    return_type = Type::tvoid;
		    break;
		default:
		    assert(0);
		}
	    }
	    break;
	case LIBCALL_ARRAYCAST:
	    t = Type::tvoid->arrayOf();
	    arg_types.push(Type::tuns32);
	    arg_types.push(Type::tuns32);
	    arg_types.push(t);
	    return_type = t;
	    break;
	case LIBCALL_ARRAYCAST_FROMBIT:
	    t = Type::tvoid->arrayOf();
	    arg_types.push(Type::tuns32);
	    arg_types.push(t);
	    return_type = t;
	    break;
	case LIBCALL_ARRAYCOPY:
	    t = Type::tvoid->arrayOf();
	    arg_types.push(Type::tuns32);
	    arg_types.push(t);
	    arg_types.push(t);
	    return_type = t;
	    break;
	case LIBCALL_ARRAYCOPYBIT:
	case LIBCALL_ARRAYCATB:
	    t = Type::tbit->arrayOf();
	    arg_types.push(t);
	    arg_types.push(t);
	    return_type = t;
	    break;
	case LIBCALL_ARRAYCAT:
	    t = Type::tvoid->arrayOf();
	    arg_types.push(t);
	    arg_types.push(t);
	    arg_types.push(Type::tuns32);
	    return_type = t;
	    break;
	case LIBCALL_ARRAYCATN:
	    arg_types.push(Type::tuns32);
	    arg_types.push(Type::tuns32);
	    varargs = true;
	    return_type = Type::tvoid->arrayOf();
	    break;
	case LIBCALL_ARRAYAPPEND:
	    t = Type::tuns8->arrayOf();
	    arg_types.push(t->pointerTo());
	    arg_types.push(t);
	    arg_types.push(Type::tuns32);
	    return_type = Type::tvoid->arrayOf();
	    break;
	case LIBCALL_ARRAYAPPENDB:
	    t = Type::tbit->arrayOf();
	    arg_types.push(t->pointerTo());
	    arg_types.push(t);
	    return_type = t;
	    break;
	case LIBCALL_ARRAYAPPENDC:
	    t = Type::tuns8->arrayOf()->pointerTo();
	    arg_types.push(t->pointerTo());
	    arg_types.push(Type::tuns32);
	    varargs = true;
	    return_type = Type::tvoid->arrayOf();
	    break;
	case LIBCALL_ARRAYAPPENDCB:
	    t = Type::tbit->arrayOf();
	    arg_types.push(t->pointerTo()); // declared inout
	    arg_types.push(Type::tbit);
	    return_type = t;
	    break;
	case LIBCALL_ARRAYSETBIT:
	    t = Type::tbit->arrayOf();
	    arg_types.push(t);
	    arg_types.push(Type::tuns32);
	    arg_types.push(Type::tuns32);
	    arg_types.push(Type::tbit);
	    return_type = t;
	    break;
	case LIBCALL_ARRAYSETBIT2:
	    t = Type::tbit->arrayOf();
	    arg_types.push(t);
	    arg_types.push(Type::tbit);
	    return_type = t;
	    break;
	case LIBCALL_MONITORENTER:
	case LIBCALL_MONITOREXIT:
	case LIBCALL_THROW:
	case LIBCALL_INVARIANT:
	    arg_types.push(getObjectType());
	    break;
	case LIBCALL_CRITICALENTER:
	case LIBCALL_CRITICALEXIT:
	    arg_types.push(Type::tvoid->pointerTo());
	    break;
	case LIBCALL_SWITCH_USTRING:
	    t = Type::twchar;
	    goto do_switch_string;
	case LIBCALL_SWITCH_DSTRING:
	    t = Type::tdchar;
	    goto do_switch_string;
	case LIBCALL_SWITCH_STRING:
	    t = Type::tchar;
	do_switch_string:
	    t = t->arrayOf();
	    arg_types.push(t->arrayOf());
	    arg_types.push(t);
	    return_type = Type::tint32;
	    break;
	default:
	    abort();
	}
	decl = FuncDeclaration::genCfunc(return_type, (char *) libcall_ids[lib_call]);
	// %% have to do this because we don't declare arguments
	{
	    TypeFunction * tf = (TypeFunction *) decl->type;
	    tf->varargs = varargs ? 1 : 0;
	    Array * args = new Array;
	    args->setDim( arg_types.dim );
	    for (unsigned i = 0; i < arg_types.dim; i++)
		args->data[i] = new Argument( In, (Type *) arg_types.data[i],
		    NULL, NULL);
	    tf->arguments = args;
	}
	libcall_decls[lib_call] = decl;
    }
    return decl;
}

tree
IRState::libCall(LibCall lib_call, Array * args, tree result_type)
{
    FuncDeclaration * func_decl = getLibCallDecl(lib_call);

    tree result =  call(func_decl, args);
    if (result_type != NULL_TREE) {
	// %% assumes caller knows what it is doing
	TREE_TYPE( result ) = result_type;
    }
    return result;
}

tree IRState::maybeExpandSpecialCall(tree call_exp)
{
    // More code duplication from C

    tree callee = TREE_OPERAND(call_exp, 0);
    tree t, op;
    if (POINTER_TYPE_P(TREE_TYPE( callee ))) {
	callee = TREE_OPERAND(callee, 0);
    }
    if (TREE_CODE(callee) == FUNCTION_DECL) {
	if (DECL_BUILT_IN_CLASS(callee) == NOT_BUILT_IN) {
	    // the most common case
	    return call_exp;
	} else if (DECL_BUILT_IN_CLASS(callee) == BUILT_IN_NORMAL) {
	    switch (DECL_FUNCTION_CODE(callee)) {
	    case BUILT_IN_ABS:
	    case BUILT_IN_LABS:
	    case BUILT_IN_LLABS:
	    case BUILT_IN_IMAXABS:
		/* The above are required for both gcc3.3 and 3.4.  The below
		   are only needed for gcc3.3, but we might as well use it
		   for gcc3.4 as well. */
	    case BUILT_IN_FABS:
	    case BUILT_IN_FABSL:
	    case BUILT_IN_FABSF:
		op = TREE_VALUE( TREE_OPERAND( call_exp, 1 ));
		t = build1(ABS_EXPR, TREE_TYPE(op), op);
		return d_convert_basic(TREE_TYPE(call_exp), t);
		// probably need a few more cases:
	    default:
		return call_exp;
	    }
	} else if (DECL_BUILT_IN_CLASS(callee) == BUILT_IN_FRONTEND) {
	    Intrinsic intrinsic = (Intrinsic) DECL_FUNCTION_CODE(callee);
	    switch (intrinsic) {
	    case INTRINSIC_C_VA_ARG:
		// %% should_check c_promotes_to as in va_arg now
		// just drop though for now...
	    case INTRINSIC_STD_VA_ARG:
		t = TREE_VALUE(TREE_OPERAND(call_exp, 1));
		// signature is (inout va_list), but VA_ARG_EXPR expects the
		// list itself...
		if ( TREE_CODE( t ) == ADDR_EXPR ) {
		    t = TREE_OPERAND(t, 0);
		} else {
		    // this probably doesn't happen... passing an inout va_list argument,
		    // but again,  it's probably { & ( * inout_arg ) }
		    t = build1(INDIRECT_REF, TREE_TYPE(TREE_TYPE(t)), t);
		}
		return build1(VA_ARG_EXPR, TREE_TYPE(TREE_TYPE(callee)), t);
	    case INTRINSIC_C_VA_START:
		/*
		t = TREE_VALUE();
		// signature is (inout va_list), but VA_ARG_EXPR expects the
		// list itself...
		if ( TREE_CODE( t ) == ADDR_EXPR ) {
		    t = TREE_OPERAND(t, 0);
		} else {
		    // this probably doesn't happen... passing an inout va_list argument,
		    // but again,  it's probably { & ( * inout_arg ) }
		    t = build1(INDIRECT_REF, TREE_TYPE(TREE_TYPE(t)), t);
		}
		*/
		// The va_list argument should already have its
		// address taken.  The second argument, however, is
		// inout and that needs to be fixed to prevent a warning.
		t = TREE_VALUE(TREE_CHAIN(TREE_OPERAND(call_exp, 1)));
		if ( TREE_CODE( t ) == ADDR_EXPR ) {
		    t = TREE_OPERAND(t, 0);
		}		
		
		return buildCall(void_type_node, // assuming nobody tries to change the return type
		    addressOf( built_in_decls[BUILT_IN_VA_START] ),
		    tree_cons(NULL_TREE, TREE_VALUE(TREE_OPERAND(call_exp, 1)),
			tree_cons( NULL_TREE, t, NULL_TREE)));
	    default:
		abort();
		break;
	    }
	    
	} else if (0 /** WIP **/ && DECL_BUILT_IN_CLASS(callee) == BUILT_IN_FRONTEND) {
	    // %%TODO: need to handle BITS_BIG_ENDIAN
	    // %%TODO: need to make expressions unsigned
	    
	    Intrinsic intrinsic = (Intrinsic) DECL_FUNCTION_CODE(callee);
	    // Might as well do intrinsics here...
	    switch (intrinsic) {
	    case INTRINSIC_BSF: // This will use bsf on x86, but BSR becomes 31-(bsf)!!
		// drop through
	    case INTRINSIC_BSR:
		// %% types should be correct, but should still check..
		// %% 64-bit..
		return call_exp;
#if D_GCC_VER >= 34
		return build(CALL_EXPR, TREE_TYPE(call_exp),
		    built_in_decls[intrinsic == INTRINSIC_BSF ? BUILT_IN_CTZ : BUILT_IN_CLZ],
		    TREE_OPERAND(call_exp, 1));
#else
		return call_exp;
#endif
	    case INTRINSIC_BT:
	    case INTRINSIC_BTC:
	    case INTRINSIC_BTR:
	    case INTRINSIC_BTS:
		break;
	    case INTRINSIC_BSWAP:
#if defined(TARGET_386)
		
#endif
		break;
	    case INTRINSIC_INP:
	    case INTRINSIC_INPW:
	    case INTRINSIC_INPL:
	    case INTRINSIC_OUTP:
	    case INTRINSIC_OUTPW:
	    case INTRINSIC_OUTPL:
#ifdef TARGET_386
#else
		::error("Port I/O intrinsic '%s' is only available on ix86 targets",
		    IDENTIFIER_POINTER(DECL_NAME(callee)));
#endif
		break;
	    default:
		abort();
	    }
	}
    }

    return call_exp;
}


#if D_GCC_VER >= 34
#ifdef ASM_OUTPUT_DEF
static int thunk_labelno;

/* Create a static alias to function.  */

static tree
make_alias_for_thunk (tree function)
{
  tree alias;
  char buf[256];

#if defined (TARGET_IS_PE_COFF)
    // make_alias_for_thunk may be be causing assembler problems on PE_COFF...
    //  if (DECL_ONE_ONLY (function)) // ... disabled
    return function;
#endif

  ASM_GENERATE_INTERNAL_LABEL (buf, "LTHUNK", thunk_labelno);
  thunk_labelno++;
  alias = build_decl (FUNCTION_DECL, get_identifier (buf),
		      TREE_TYPE (function));
  //DECL_LANG_SPECIFIC (alias) = DECL_LANG_SPECIFIC (function);
  //cxx_dup_lang_specific_decl (alias);
  DECL_CONTEXT (alias) = NULL;
  TREE_READONLY (alias) = TREE_READONLY (function);
  TREE_THIS_VOLATILE (alias) = TREE_THIS_VOLATILE (function);
  TREE_PUBLIC (alias) = 0;
  //DECL_INTERFACE_KNOWN (alias) = 1;
  //DECL_NOT_REALLY_EXTERN (alias) = 1;
  //DECL_THIS_STATIC (alias) = 1;
  //DECL_SAVED_FUNCTION_DATA (alias) = NULL;
  //DECL_DESTRUCTOR_P (alias) = 0;
  //DECL_CONSTRUCTOR_P (alias) = 0;
  //DECL_CLONED_FUNCTION (alias) = NULL_TREE;
  DECL_EXTERNAL (alias) = 0;
  DECL_ARTIFICIAL (alias) = 1;
  DECL_NO_STATIC_CHAIN (alias) = 1;
  //DECL_PENDING_INLINE_P (alias) = 0;
  DECL_INLINE (alias) = 0;
  DECL_DECLARED_INLINE_P (alias) = 0;
  //DECL_DEFERRED_FN (alias) = 0;
  //DECL_USE_TEMPLATE (alias) = 0;
  //DECL_TEMPLATE_INSTANTIATED (alias) = 0;
  //DECL_TEMPLATE_INFO (alias) = NULL;
  //DECL_INITIAL (alias) = error_mark_node;
  TREE_ADDRESSABLE (alias) = 1;
  TREE_USED (alias) = 1;
  SET_DECL_ASSEMBLER_NAME (alias, DECL_NAME (alias));
  TREE_SYMBOL_REFERENCED (DECL_ASSEMBLER_NAME (alias)) = 1;
  if (!flag_syntax_only)
    assemble_alias (alias, DECL_ASSEMBLER_NAME (function));
  return alias;
}
#endif
#endif

// Can't output thunks while a function is being compiled. Called from
// d-decls.cc (no cur_irs)
static Array deferredThunks;

typedef struct {
    tree decl;
    tree target;
    int offset;
} DeferredThunk;

void
IRState::doThunk(tree thunk_decl, tree target_decl, int offset)
{
    if (current_function_decl) {
	DeferredThunk * t = new DeferredThunk;
	t->decl = thunk_decl;
	t->target = target_decl;
	t->offset = offset;
	deferredThunks.push(t);
    } else {
	outputThunk(thunk_decl, target_decl, offset);
    }
}

IRState *
IRState::endFunction()
{
    IRState * irs = IRBase::endFunction();

    if ( ! current_function_decl ) { // Could just hold off to the end of the module...
		for (unsigned i = 0; i < deferredThunks.dim; i++) {
	    DeferredThunk * t = (DeferredThunk *) deferredThunks.data[i];
	    outputThunk(t->decl, t->target, t->offset);
	}
	deferredThunks.setDim(0);
    }

    return irs;
}

void
IRState::outputThunk(tree thunk_decl, tree target_decl, int offset)
{
    int delta = -offset;
    tree alias;
#if D_GCC_VER >= 34
#ifdef ASM_OUTPUT_DEF
    alias = make_alias_for_thunk(target_decl);
#else
    alias = target_decl;
#endif
#else
    alias = target_decl;
#endif

    TREE_ADDRESSABLE(target_decl) = 1;
    TREE_USED(target_decl) = 1;
#if D_GCC_VER >= 34
    DECL_VISIBILITY (thunk_decl) = DECL_VISIBILITY (target_decl);
#if defined (ASM_OUTPUT_DEF) \
  && !defined (TARGET_IS_PE_COFF)
  if (targetm.have_named_sections)
    {
      resolve_unique_section (target_decl, 0, flag_function_sections);

      if (DECL_SECTION_NAME (target_decl) != NULL && DECL_ONE_ONLY (target_decl))
	{
	  resolve_unique_section (thunk_decl, 0, flag_function_sections);

	  /* Output the thunk into the same section as function.  */
	  DECL_SECTION_NAME (thunk_decl) = DECL_SECTION_NAME (target_decl);
	}
    }
#endif
#endif
  // cp/method.c:
  /* The back-end expects DECL_INITIAL to contain a BLOCK, so we
     create one.  */
  // ... actually doesn't seem to be the case for output_mi_thunk
  DECL_INITIAL (thunk_decl) = make_node (BLOCK);
  BLOCK_VARS (DECL_INITIAL (thunk_decl)) = DECL_ARGUMENTS (thunk_decl);

    if (targetm.asm_out.can_output_mi_thunk(thunk_decl, delta, 0, alias)) {
	const char *fnname;

	current_function_decl = thunk_decl;
	DECL_RESULT( thunk_decl ) = build_decl(RESULT_DECL, 0, integer_type_node);
	
	fnname = XSTR(XEXP(DECL_RTL(thunk_decl), 0), 0);
	initFunctionStart(thunk_decl, 0);
	current_function_is_thunk = 1;
	assemble_start_function (thunk_decl, fnname);
	targetm.asm_out.output_mi_thunk (asm_out_file, thunk_decl,
	    delta, 0, alias);
	assemble_end_function(thunk_decl, fnname);
	current_function_decl = 0;
	cfun = 0;
#if D_GCC_VER < 40
	/* Because init_function_start increments this, we must
	   decrement it.  */
	immediate_size_expand--;
#endif
	TREE_ASM_WRITTEN (thunk_decl) = 1;
    } else {
	sorry("backend for this target machine does not support thunks");
    }
}


bool
IRState::functionNeedsChain(FuncDeclaration *f)
{
    Dsymbol * s;
    ClassDeclaration * a;
    
    if (f->isNested())
	return true;

    s = f->toParent();

    while ( s && (a = s->isClassDeclaration()) && a->isNested() ) {
	s = s->toParent();
	if (s->isFuncDeclaration())
	    return true;
    }
    return false;
}

tree
IRState::doSimpleFunction(Module * mod, const char * name, tree expr, bool static_ctor, bool public_fn)
{
    TypeFunction * func_type = new TypeFunction(0, Type::tvoid, 0, LINKc);
    FuncDeclaration * func = new FuncDeclaration(mod->loc, mod->loc, // %% locs may be wrong
	Lexer::idPool(name), STCstatic, func_type); // name is only to prevent crashes
    func->loc = Loc(mod, 1); // to prevent debug info crash // maybe not needed if DECL_ARTIFICIAL?
    func->linkage = func_type->linkage;
    func->parent = mod;
    func->protection = PROTprivate;

    tree func_decl = func->toSymbol()->Stree;
    if (static_ctor) {
	DECL_STATIC_CONSTRUCTOR( func_decl ) = 1; // apparently, the back end doesn't do anything with this
	TREE_USED( func_decl ) = 1; // only needed for GCC 3.3?
    }
    // D static ctors, dtors, unittests, and the ModuleInfo chain function
    // are always private (see IRState::setupSymbolStorage, default case)
    TREE_PUBLIC( func_decl ) = public_fn;
    
    // %% maybe remove the identifier

    func->fbody = new ExpStatement(mod->loc,
	new WrappedExp(mod->loc, TOKcomma, expr, Type::tvoid));

    func->toObjFile();

    return addressOf(func_decl);
}

tree
IRState::doFunctionToCallFunctions(Module * mod, const char * name, Array * functions)
{
    // If there is only one function, just return that
    if (functions->dim == 1) {
	return addressOf( (Declaration *) functions->data[0] );
    } else if (functions->dim > 1) {
	// %% shouldn't front end build these?
	tree exp = NULL_TREE;
	tree call_exp = NULL_TREE;
	for (unsigned i = 0; i < functions->dim; i++) {
	    FuncDeclaration * fn_decl = (FuncDeclaration *) functions->data[i];
	    call_exp = gen.buildCall(void_type_node, gen.addressOf(fn_decl), NULL_TREE);
	    if (! exp)
		exp = call_exp;
	    else {
		exp = build(COMPOUND_EXPR, void_type_node, exp, call_exp);
	    }
	}
	Symbol * s = mod->toSymbolX(name, 0, 0);
	return gen.doSimpleFunction(mod, s->Sident, exp, false);
    } else {
	return NULL_TREE;
    }
}


/* Type of return value may be different from aer_exp->exp.
   Case 1: For array of bit, type is bitConfig.elemType
*/

tree
IRState::arrayElemRef(IndexExp * aer_exp, ArrayScope * aryscp,
    tree * and_mask_out, tree * or_mask_out, tree * shift_count_out)
{
    if (and_mask_out)
	*and_mask_out = NULL_TREE;
    if (or_mask_out)
	*or_mask_out = NULL_TREE;
    if (shift_count_out)
	*shift_count_out = NULL_TREE;
    
    Expression * e1 = aer_exp->e1;
    Expression * e2 = aer_exp->e2;
	
    Type * base_type = e1->type->toBasetype();
    TY base_type_ty = base_type->ty;

    bool is_bit = base_type->next->isbit();

    tree index_expr; // logical index
    tree subscript_expr; // expr that indexes the array data; different from index_expr for bit for bounds check
    tree ptr_exp;  // base pointer to the elements
    tree elem_ref; // reference the the element.  built in two steps for bit

    index_expr = e2->toElem( this );
    subscript_expr = index_expr;

    switch (base_type_ty) {
    case Tarray:
    case Tsarray:
	{
	    tree e1_tree = e1->toElem(this);

	    e1_tree = aryscp->setArrayExp(e1_tree, e1->type);

	    if ( global.params.useArrayBounds &&
		// If it's a static array and the index is
		// constant, the front end has already
		// checked the bounds.
		! (base_type_ty == Tsarray && e2->isConst()) ) {

		
		tree array_len_expr, throw_expr, oob_cond;
		// implement bounds check as a conditional expression:
		// a[ inbounds(index) ? index : { throw ArrayBoundsError } ]
		//
		// First, set up the index expression to only be evaluated
		// once.
		// %% save_expr does this check: if (! TREE_CONSTANT( index_expr ))
		//   %% so we don't do a <0 check for a[2]...
		index_expr = maybeMakeTemp( index_expr );
		
		if (base_type_ty == Tarray) {
		    e1_tree = maybeMakeTemp(e1_tree);
		    array_len_expr = darrayLenRef(e1_tree);
		} else {
		    array_len_expr = ((TypeSArray *) base_type)->dim->toElem(this);
		}

		oob_cond = boundsCond(index_expr, array_len_expr, false);
		throw_expr = assertCall(aer_exp->loc, LIBCALL_ARRAY_BOUNDS);

		subscript_expr = build( COND_EXPR, TREE_TYPE( index_expr ),
		    oob_cond, index_expr, throw_expr );
	    }

	    if (is_bit) {
		// %% dup'd below for Tpointer case
		// %% do we want the orig index_expr or the subscript_expr saved?
		
		index_expr = maybeMakeTemp( index_expr );
		
		tree shift_node = bitConfig.shift_tree;
		// Type::tindex is signed, subscript_expr must be unsigned to generatate unsigned
		// right shift %% 
		//if ( ! TYPE_UNSIGNED( TREE_TYPE( subscript_expr )) )..
		// %% YOW! calling 'convert' breaks stuff
		subscript_expr = d_convert_basic( TREE_TYPE( shift_node ), subscript_expr ); // folding?
		subscript_expr = build( RSHIFT_EXPR, TREE_TYPE(shift_node),
		    subscript_expr, shift_node );

		subscript_expr = fold( subscript_expr ); // %% won't work for bound check..
		// would need COMPOUND_EXPR( bounds_check_expr, folded_index_expr )
		
	    }

	    // %% TODO: make this an ARRAY_REF?
	    if (base_type_ty == Tarray)
		ptr_exp = darrayPtrRef( e1_tree ); // %% do convert in darrayPtrRef?
	    else
		ptr_exp = addressOf( e1_tree );
	    // This conversion is required for static arrays and is just-to-be-safe
	    // for dynamic arrays
	    ptr_exp = d_convert_basic(base_type->next->pointerTo()->toCtype(), ptr_exp);
	}
	break;
    case Tpointer:
	// Ignores aryscp
	if (is_bit) {
	    // %% dup'd above
	    if ( TREE_CODE( index_expr ) != SAVE_EXPR)
		index_expr = save_expr( index_expr );
	    subscript_expr = index_expr;
	    if ( ! TREE_UNSIGNED(TREE_TYPE(subscript_expr)))
		subscript_expr = d_convert_basic( d_unsigned_type(TREE_TYPE(subscript_expr)),
		    subscript_expr);
	    subscript_expr = build(RSHIFT_EXPR, TREE_TYPE(subscript_expr),
		subscript_expr,
		bitConfig.shift_tree/*integerConstant(bitarray_elem_shift, unsigned_type_node)*/);
	    subscript_expr = fold( subscript_expr );
	}
	ptr_exp = e1->toElem( this );
	break;
    default:
	abort();
    }

    if (! is_bit)
	ptr_exp = pvoidOkay( ptr_exp );
    else
	ptr_exp = d_convert_basic(bitConfig.elemType->pointerTo()->toCtype(), ptr_exp);
    elem_ref = build1(INDIRECT_REF, TREE_TYPE(TREE_TYPE(ptr_exp)),
	pointerIntSum( ptr_exp, subscript_expr));

    if (is_bit) {
	tree t;

	// (index & <mask>)
	tree bit_index_expr = build(BIT_AND_EXPR, TREE_TYPE(index_expr),
	    index_expr, bitConfig.mask_tree/*integerConstant(bitarray_elem_mask, bitArrayElemType)*/);
	bit_index_expr = fold( bit_index_expr );

	if (or_mask_out || shift_count_out)
	    bit_index_expr = save_expr( bit_index_expr );
	// 1 << (index & <mask>)
	t = build(LSHIFT_EXPR, bitConfig.elemType->toCtype()/*bitArrayElemType->toCtype()*/,
	    integer_one_node, bit_index_expr);
	if (and_mask_out)
	    * and_mask_out = fold(t);

	if (shift_count_out)
	    * shift_count_out = bit_index_expr;
    }

    return elem_ref;
}

tree
IRState::arrayElemRef(IndexExp * aer_exp, ArrayScope * aryscp)
{
    tree elem_ref, and_mask;
    elem_ref = arrayElemRef(aer_exp, aryscp, & and_mask, (tree*) 0, (tree*) 0);
    if (and_mask) { // i.e., is the element type bit.
	/*
	// %% d_truthvalue_conversion will convert something that is already truthval cvt'd? .. would
	// be nice to optimize this out...
	// %% also assumptions about boolean_type_node == Type::tboolean
	if (TREE_CODE(and_mask) == BIT_NOT_EXPR)
	    and_mask = TREE_OPERAND( and_mask, 0 );
	else
	    and_mask = fold( build1(BIT_NOT_EXPR, TREE_TYPE(and_mask), and_mask) );
	*/
	return d_truthvalue_conversion( build(BIT_AND_EXPR,
				TREE_TYPE(elem_ref), elem_ref, and_mask ));
    } else
	return elem_ref;
}

tree
IRState::darrayPtrRef(tree exp)
{
    // Get the backend type for the array and pick out the array data
    // pointer field (assumed to be the second field.)
    tree ptr_field = TREE_CHAIN( TYPE_FIELDS( TREE_TYPE( exp )));
    return build(COMPONENT_REF, TREE_TYPE( ptr_field ), exp, ptr_field);
}

tree
IRState::darrayLenRef(tree exp)
{
    // Get the backend type for the array and pick out the array length
    // field (assumed to be the first field.)
    tree len_field = TYPE_FIELDS( TREE_TYPE( exp ));
    return build(COMPONENT_REF, TREE_TYPE( len_field ), exp, len_field);
}


tree
IRState::darrayVal(tree type, tree len, tree data)
{
    // %% assert type is a darray
    tree ctor = make_node( CONSTRUCTOR );
    tree len_field, ptr_field;
    tree elts;
    
    TREE_TYPE( ctor ) = type;
    TREE_READONLY( ctor ) = 1;
    TREE_STATIC( ctor ) = 0;   // can be set by caller if needed
    TREE_CONSTANT( ctor ) = 0; // "
    len_field = TYPE_FIELDS( TREE_TYPE( ctor ));
    ptr_field = TREE_CHAIN( len_field );

    elts = tree_cons( ptr_field, data, NULL_TREE ); // shouldn't need to convert the pointer...
    elts = tree_cons( len_field, len, elts );
    CONSTRUCTOR_ELTS( ctor ) = elts;

    return ctor;
}

tree
IRState::darrayVal(tree type, unsigned len, tree data)
{
    // %% assert type is a darray
    tree ctor = make_node( CONSTRUCTOR );
    tree len_value, ptr_value, len_field, ptr_field;
    tree elts;
    
    TREE_TYPE( ctor ) = type;
    TREE_READONLY( ctor ) = 1;
    TREE_STATIC( ctor ) = 0;   // can be set by caller if needed
    TREE_CONSTANT( ctor ) = 0; // "
    len_field = TYPE_FIELDS( TREE_TYPE( ctor ));
    ptr_field = TREE_CHAIN( len_field );

    if (data ) {
	assert( POINTER_TYPE_P( TREE_TYPE( data )));
	ptr_value = data;
    } else {
	ptr_value = convert( TREE_TYPE( ptr_field ), integer_zero_node);
    }

    elts = tree_cons( ptr_field, ptr_value, NULL_TREE ); // shouldn't need to convert the pointer...
    
    len_value = integerConstant(len, TREE_TYPE(len_field));
    elts = tree_cons( len_field, len_value, elts );
    CONSTRUCTOR_ELTS( ctor ) = elts;

    return ctor;
}

tree
IRState::darrayString(const char * str)
{
    unsigned len = strlen(str);
    // %% assumes str is null-terminated
    tree str_tree = build_string(len + 1, str);
    
    TREE_TYPE( str_tree ) = arrayType(Type::tchar, len);
    return darrayVal(Type::tchar->arrayOf()->toCtype(), len, addressOf(str_tree));
}

char *
IRState::hostToTargetString(char * str, unsigned length, unsigned unit_size)
{
    if (unit_size == 1)
	return str;
    assert(unit_size == 2 || unit_size == 4);
    
    bool flip;
#ifdef HOST_WORDS_BIG_ENDIAN
    flip = (bool) ! BYTES_BIG_ENDIAN;
#else
    flip = (bool) BYTES_BIG_ENDIAN;
#endif

    if (flip) {
	char * out_str = (char *) xmalloc(length * unit_size);
	const d_uns8 * p_src = (const d_uns8 *) str;
	d_uns8 * p_out = (d_uns8 *) out_str;

	while (length--) {
	    if (unit_size == 2) {
		p_out[0] = p_src[1];
		p_out[1] = p_src[0];
	    } else /* unit_size == 4 */ {
		p_out[0] = p_src[3];
		p_out[1] = p_src[2];
		p_out[2] = p_src[1];
		p_out[3] = p_src[0];
	    }
	    p_src += unit_size;
	    p_out += unit_size;
	}
	return out_str;
    } else {
	return str;
    }
}


tree
IRState::arrayLength(tree exp, Type * exp_type)
{
    Type * base_type = exp_type->toBasetype();
    switch (base_type->ty) {
    case Tsarray:
	return size_int( ((TypeSArray *) base_type)->dim->toUInteger() );
    case Tarray:
	return darrayLenRef(exp);
    default:
	::error("can't determine the length of a %s", exp_type->toChars());
	return error_mark_node;
    }
}
    
tree
IRState::floatMod(tree a, tree b, Type * d_type)
{
#if D_GCC_VER >= 34
    enum built_in_function fn;
    switch (d_type->toBasetype()->ty) {
    case Tfloat32:
    case Timaginary32:
	fn = BUILT_IN_FMODF;
	break;
    case Tfloat64:
    case Timaginary64:
    no_long_double:
	fn = BUILT_IN_FMOD;
	break;
    case Tfloat80:
    case Timaginary80:
	if (! haveLongDouble())
	    goto no_long_double;
	fn = BUILT_IN_FMODL;
	break;
    default:
	::error("tried to perform floating-point modulo division on %s",
	    d_type->toChars());
	return error_mark_node;
    }
    tree decl = built_in_decls[fn];
    // %% assuming no arg conversion needed
    // %% bypassing buildCall since this shouldn't have
    // side effects
    return build(CALL_EXPR, TREE_TYPE(TREE_TYPE(decl)),
	addressOf(decl),
	tree_cons(NULL_TREE, a,
	    tree_cons(NULL_TREE, b, NULL_TREE)));
#else
    // could move this to d-builtins.c
    static tree float_mod_decls[3];
    static const char * decl_names[3] = { "fmodf", "fmod", "fmodl" };
    tree the_float_type;
    int decl_idx;
    switch (d_type->toBasetype()->ty) {
    case Tfloat32:
    case Timaginary32:
	decl_idx = 0;
	the_float_type = float_type_node;
	break;
    case Tfloat64:
    case Timaginary64:
    no_long_double:
	decl_idx = 1;
	the_float_type = double_type_node;
	break;
    case Tfloat80:
    case Timaginary80:
	if (! haveLongDouble())
	    goto no_long_double;
	decl_idx = 2;
	the_float_type = long_double_type_node;
	break;
    default:
	::error("tried to perform floating-point modulo division on %s",
	    d_type->toChars());
	return error_mark_node;
    }
    
    tree decl = float_mod_decls[decl_idx];
    if (! decl) {
	tree func_type = build_function_type(the_float_type,
	    tree_cons(NULL_TREE, the_float_type,
		tree_cons(NULL_TREE, the_float_type,
		    tree_cons(NULL_TREE, void_type_node, NULL_TREE))));
	decl = build_decl(FUNCTION_DECL, get_identifier(decl_names[decl_idx]),
	    func_type);
	DECL_EXTERNAL(decl) = 1;
	float_mod_decls[decl_idx] = decl;
    }
    // %% assuming no arg conversion needed
    // %% bypassing buildCall since this shouldn't have
    // side effects
    return build(CALL_EXPR, TREE_TYPE(TREE_TYPE(decl)),
	addressOf(decl),
	tree_cons(NULL_TREE, a,
	    tree_cons(NULL_TREE, b, NULL_TREE)));
#endif
}

tree
IRState::typeinfoReference(Type * t)
{
    tree ti_ref = t->getInternalTypeInfo(NULL)->toElem(this);
    assert( POINTER_TYPE_P( TREE_TYPE( ti_ref )) );
    return ti_ref;
}


// delegate is
// struct delegate {
//   void * frame_or_object;
//   void * function;
// }

tree
IRState::delegateObjectRef(tree exp)
{
    // Get the backend type for the array and pick out the array data
    // pointer field (assumed to be the first field.)
    tree obj_field = TYPE_FIELDS( TREE_TYPE( exp ));
    return build(COMPONENT_REF, TREE_TYPE( obj_field ), exp, obj_field);
}

tree
IRState::delegateMethodRef(tree exp)
{
    // Get the backend type for the array and pick out the array length
    // field (assumed to be the second field.)
    tree method_field = TREE_CHAIN( TYPE_FIELDS( TREE_TYPE( exp )));
    return build(COMPONENT_REF, TREE_TYPE( method_field ), exp, method_field);
}

// Converts pointer types of method_exp and object_exp to match d_type
tree
IRState::delegateVal(tree method_exp, tree object_exp, Type * d_type)
{
    Type * base_type = d_type->toBasetype();
    if ( base_type->ty == Tfunction ) {
	// Called from DotVarExp.  These are just used to
	// make function calls and not to make Tdelegate variables.
	// Clearing the type makes sure of this.
	base_type = 0;
    } else {
	assert(base_type->ty == Tdelegate);
    }
    
    tree type = base_type ? base_type->toCtype() : NULL_TREE;
    tree ctor = make_node( CONSTRUCTOR );
    tree obj_field = NULL_TREE;
    tree func_field = NULL_TREE;
    
    TREE_READONLY( ctor ) = 1;
    if (type) {
	TREE_TYPE( ctor ) = type;
	obj_field = TYPE_FIELDS( type );
	func_field = TREE_CHAIN( obj_field );
    }
    CONSTRUCTOR_ELTS( ctor ) =
	tree_cons( obj_field,  object_exp,
	    tree_cons( func_field, method_exp, NULL));

    return ctor;
}

void
IRState::extractMethodCallExpr(tree mcr, tree & callee_out, tree & object_out) {
    assert( D_IS_METHOD_CALL_EXPR( mcr ));
    tree elts = CONSTRUCTOR_ELTS( mcr );
    object_out = TREE_VALUE( elts );
    callee_out = TREE_VALUE( TREE_CHAIN( elts ));
}

tree
IRState::objectInstanceMethod(Expression * obj_exp, FuncDeclaration * func, Type * d_type)
{
    Type * obj_type = obj_exp->type->toBasetype();
    if (func->isThis()) {
	bool is_dottype;
	tree this_expr;

	// DotTypeExp cannot be evaluated
	if (obj_exp->op == TOKdottype) {
	    is_dottype = true;
	    this_expr = ((DotTypeExp *) obj_exp)->e1->toElem( this );
	} else if (obj_exp->op == TOKcast &&
	    ((CastExp*) obj_exp)->e1->op == TOKdottype) {
	    is_dottype = true;
	    // see expression.c:"See if we need to adjust the 'this' pointer"
	    this_expr = ((DotTypeExp *) ((CastExp*) obj_exp)->e1)->e1->toElem( this );
	} else {
	    is_dottype = false;
	    this_expr = obj_exp->toElem( this );
	}
	
	// Calls to super are static (func is the super's method)
	// Structs don't have vtables.
	// Final and non-virtual methods can be called directly.
	// DotTypeExp means non-virtual
	
	if (obj_exp->op == TOKsuper ||
	    obj_type->ty == Tstruct || obj_type->ty == Tpointer ||
	    func->isFinal() || ! func->isVirtual() || is_dottype) {
	    
	    if (obj_type->ty == Tstruct)
		this_expr = addressOf(this_expr);
	    return methodCallExpr(functionPointer(func), this_expr, d_type);
	} else {
	    // Interface methods are also in the class's vtable, so we don't
	    // need to convert from a class pointer to an interface pointer.
	    this_expr = maybeMakeTemp( this_expr );

	    tree vtbl_ref = build1( INDIRECT_REF, TREE_TYPE(TREE_TYPE( this_expr )),
		this_expr ); // dereference the pointer to the object
	    tree field = TYPE_FIELDS( TREE_TYPE( vtbl_ref )); // the vtbl is the first field
	    vtbl_ref = build( COMPONENT_REF, TREE_TYPE( field ), vtbl_ref, field ); // vtbl field (a pointer)
	    // %% better to do with array ref?
	    vtbl_ref = build( PLUS_EXPR, TREE_TYPE(vtbl_ref), vtbl_ref,
		size_int( getPointerSize() * func->vtblIndex ));
	    vtbl_ref = build1( INDIRECT_REF, TREE_TYPE( functionPointer(func) ),
		vtbl_ref );

	    return methodCallExpr(vtbl_ref, this_expr, d_type);
	}
    } else {
	// Static method; ignore the object instance
	return addressOf(func);
    }
}


tree
IRState::realPart(tree c) {
    return build1(REALPART_EXPR, TREE_TYPE(TREE_TYPE(c)), c);
}
tree
IRState::imagPart(tree c) {
    return build1(IMAGPART_EXPR, TREE_TYPE(TREE_TYPE(c)), c);
}

tree
IRState::twoFieldType(tree rec_type, tree ft1, tree ft2, Type * d_type)
{
    tree f0 = build_decl(FIELD_DECL, get_identifier("_a"), ft1);
    tree f1 = build_decl(FIELD_DECL, get_identifier("_b"), ft2);
    DECL_CONTEXT(f0) = rec_type;
    DECL_CONTEXT(f1) = rec_type;
    TYPE_FIELDS(rec_type) = chainon(f0, f1);
    layout_type(rec_type);
#ifdef ORIG
    TYPE_STUB_DECL(rec_type) = build_decl(TYPE_DECL, NULL_TREE, rec_type);
    rest_of_type_compilation(rec_type, 1);
#else
    /*
    if (d_type) {
	// EXPER
	TYPE_NAME(rec_type) = build_decl(TYPE_DECL, get_identifier(d_type->toChars()), rec_type); // this, and rest_of_decl.. is all that is needed for dbxout
	//TYPE_STUB_DECL(rec_type) = copy_node(TYPE_NAME(rec_type));
	//DECL_ARTIFICIAL(TYPE_STUB_DECL(rec_type)) = 1;
	//DECL_IGNORED_P(TYPE_STUB_DECL(rec_type)) = 1;
	Loc l;
	l.filename = "<internal>";
	l.linnum = 1;
	//setDeclLoc(TYPE_NAME(rec_type), l);
	//pushdecl(TYPE_NAME(rec_type));
	// global_binding_level->names
	//rest_of_type_compilation(rec_type, 1);
	rest_of_decl_compilation(TYPE_NAME(rec_type), NULL, 1, 0);
    }
    */
#endif
    if (d_type) {
	declareType(rec_type, d_type);
	TYPE_LANG_SPECIFIC(rec_type) = build_d_type_lang_specific(d_type);
    }
    return rec_type;
}

// Create a record type from two field types
tree
IRState::twoFieldType(Type * ft1, Type * ft2, Type * d_type)
{
    return twoFieldType( make_node( RECORD_TYPE ), ft1->toCtype(), ft2->toCtype(), d_type );
}

tree
IRState::twoFieldCtor(tree rec_type, tree f1, tree f2, int storage_class)
{
    tree ctor = make_node( CONSTRUCTOR );
    tree ft1, ft2;
    
    TREE_TYPE( ctor ) = rec_type;
    TREE_READONLY( ctor ) = 1;
    TREE_STATIC( ctor ) = (storage_class & STCstatic) != 0;   
    TREE_CONSTANT( ctor ) = (storage_class & STCconst) != 0;   
    ft1 = TYPE_FIELDS( rec_type );
    ft2 = TREE_CHAIN( ft1 );

    CONSTRUCTOR_ELTS( ctor ) = tree_cons( ft1, f1,
	tree_cons( ft2, f2, NULL_TREE ));

    return ctor;
}

// This could be made more lax to allow better CSE (?)
static bool
needs_temp(tree t) {
    // %%TODO: check for anything with TREE_SIDE_EFFECTS?
    switch (TREE_CODE(t)) {
    case VAR_DECL:
    case FUNCTION_DECL:
    case PARM_DECL:
    case CONST_DECL:
    case SAVE_EXPR:
	return false;
    case ADDR_EXPR:
    case REFERENCE_EXPR:
    case INDIRECT_REF:
    case COMPONENT_REF:
    case NOP_EXPR:
    case NON_LVALUE_EXPR:
    case VIEW_CONVERT_EXPR:
	return needs_temp(TREE_OPERAND(t, 0));
    case ARRAY_REF:
	return true;
    default:
	return true;
    }
}

Expression *
IRState::maybeMakeTemp(Expression * e)
{
    // %%TODO if no temp needed, juse return e...
    
    tree result = e->toElem( this); // %% ugly
    if (needs_temp(result))
	// save_expr can produce poorly performing code; eventually make a real temp for these cases
	// %%TODO: using e->op safe?
	result = save_expr(result);
    return new WrappedExp(e->loc, e->op, result, e->type);
}

tree
IRState::maybeMakeTemp(tree t)
{
    // save_expr can produce poorly performing code; eventually make a real temp for these cases
    // e->op ok? eventually make TOKvar for a real temp
    return needs_temp(t) ? save_expr(t) : t;
}

tree
IRState::toElemLvalue(Expression * e)
{
    /*
    if (e->op == TOKcast)
	fprintf(stderr, "IRState::toElemLvalue TOKcast\n");
    else
    */
	
    if (e->op == TOKindex) {
	IndexExp * ie = (IndexExp *) e;
	Expression * e1 = ie->e1;
	Expression * e2 = ie->e2;
	Type * type = e->type;
	Type * array_type = e1->type->toBasetype();
	
	if (array_type->ty == Taarray) {
	    Type * key_type = ((TypeAArray *) array_type)->key->toBasetype();
	    tree args[4];
	    args[0] = this->addressOf( e1->toElem(this) );
	    args[1] = this->typeinfoReference(key_type);
	    args[2] = this->integerConstant( array_type->next->size(), Type::tuns32 );
	    args[3] = this->convertTo( e2, key_type ); // %% vararg issues
	    return build1(INDIRECT_REF, type->toCtype(),
		this->libCall(LIBCALL_AAGET, 4, args, type->pointerTo()->toCtype()));
	}
    }
    return e->toElem(this);
}

tree
IRState::aggDeclStaticInit(AggregateDeclaration * agg_decl)
{
    SymbolDeclaration * symdec = new SymbolDeclaration(0, agg_decl->toInitializer(),
	agg_decl->isStructDeclaration());
    tree t = symdec->toSymbol()->Stree;
    delete symdec;
    return t;
}

Module * IRState::builtinsModule = 0;
Module * IRState::intrinsicModule = 0;
TemplateDeclaration * IRState::stdargTemplateDecl = 0;
TemplateDeclaration * IRState::cstdargStartTemplateDecl = 0;
TemplateDeclaration * IRState::cstdargArgTemplateDecl = 0;

static inline const char * strip_builtin(const char * x) {
    if (! strncmp(x, "__builtin_", strlen("__builtin_")))
	return x + strlen("__builtin_");
    else
	return x;
}

bool
IRState::maybeSetUpBuiltin(Declaration * decl)
{
    Dsymbol * dsym;
    TemplateInstance * ti;

    dsym = decl->toParent();
    if (dsym->getModule() == intrinsicModule) {
	// Matches order of Intrinsic enum
	static const char * intrinsic_names[] = {
	    "bsf", "bsr",
	    "bt", "btc", "btr", "bts",
	    "bswap",
	    "inp", "inpw", "inpl",
	    "outp", "outw", "outl", NULL
	};
	for (int i = 0; intrinsic_names[i]; i++) {
	    if ( ! strcmp( decl->ident->string, intrinsic_names[i] ) ) {
		bool have_intrinsic = false;
		tree t = decl->toSymbol()->Stree;

		switch ( (Intrinsic) i ) {
		case INTRINSIC_BSF:
		case INTRINSIC_BSR:
		case INTRINSIC_BT:
		case INTRINSIC_BTC:
		case INTRINSIC_BTR:
		case INTRINSIC_BTS:
		    break;
		case INTRINSIC_BSWAP:
#if defined(TARGET_386)
		    //have_intrinsic = true;
#endif
		    break;
		case INTRINSIC_OUTP:
		case INTRINSIC_OUTPW:
		case INTRINSIC_OUTPL:
		    // Only on ix86, but need to given error message on others
		    have_intrinsic = true;
		    break;
		default:
		    abort();
		}
		
		if (have_intrinsic) {
		    DECL_BUILT_IN_CLASS( t ) = BUILT_IN_FRONTEND;
		    DECL_FUNCTION_CODE( t ) = (built_in_function) i;
		    return true;
		} else
		    return false;
	    }
	}
    } else if (dsym) {
	ti = dsym->isTemplateInstance();
	if (ti) {
	    tree t = decl->toSymbol()->Stree;
	    if (ti->tempdecl == stdargTemplateDecl) {
		DECL_BUILT_IN_CLASS(t) = BUILT_IN_FRONTEND;
		DECL_FUNCTION_CODE(t) = (built_in_function) INTRINSIC_STD_VA_ARG;
		return true;
	    } else if (ti->tempdecl == cstdargArgTemplateDecl) {
		DECL_BUILT_IN_CLASS(t) = BUILT_IN_FRONTEND;
		DECL_FUNCTION_CODE(t) = (built_in_function) INTRINSIC_C_VA_ARG;
		return true;
	    } else if (ti->tempdecl == cstdargStartTemplateDecl) {
		DECL_BUILT_IN_CLASS(t) = BUILT_IN_FRONTEND;
		DECL_FUNCTION_CODE(t) = (built_in_function) INTRINSIC_C_VA_START;
		return true;
	    }
	}
    }
    return false;
}

bool
IRState::isDeclarationReferenceType(Declaration * decl)
{
    Type * base_type = decl->type->toBasetype();

    // D doesn't do this now..
    if ( base_type->ty == Treference ) {
	return true;
    }

    if (  decl->isOut() ||
	( decl->isParameter() && base_type->ty == Tsarray ) ) {
	return true;
    }

    return false;
}

tree
IRState::trueDeclarationType(Declaration * decl)
{
    // If D supported references, we would have to check twice for
    //   (out T &) -- disallow, maybe or make isDeclarationReferenceType return
    //   the number of levels to reference
    tree decl_type = decl->type->toCtype();
    if ( isDeclarationReferenceType( decl )) {
	return build_reference_type( decl_type );
    } else {
	return decl_type;
    }
}

// These should match the Declaration versions above
bool
IRState::isArgumentReferenceType(Argument * arg)
{
    Type * base_type = arg->type->toBasetype();

    if ( base_type->ty == Treference ) {
	return true;
    }

    if ( (arg->inout == Out) || (arg->inout == InOut) || base_type->ty == Tsarray ) {
	return true;
    }

    return false;
}

tree
IRState::trueArgumentType(Argument * arg)
{
    tree arg_type = arg->type->toCtype();
    if ( isArgumentReferenceType( arg )) {
	return build_reference_type( arg_type );
    } else {
	return arg_type;
    }
}

tree
IRState::arrayType(tree type_node, uinteger_t size)
{
    tree index_type_node;
    if (size > 0) {
	index_type_node = size_int( size - 1 );
	index_type_node = build_index_type(index_type_node);
    } else {
	// See c-decl.c grokdeclarator for zero-length arrays
	// %%TODO: does constant needs to be signed?
	index_type_node = build_range_type (sizetype, integerConstant(size, sizetype), NULL_TREE);
    }
    return build_array_type(type_node, index_type_node);
}

tree
IRState::addTypeAttribute(tree type, const char * attrname, tree value)
{
    // use build_type_copy / build_type_attribute_variant
    
    // types built by functions in tree.c need to be treated as immutable
    if ( ! TYPE_ATTRIBUTES( type )) { // ! TYPE_ATTRIBUTES -- need a better check
	type = build_type_copy( type );
	// TYPE_STUB_DECL( type ) = .. if we need this for structs, etc.. since
	// TYPE_STUB_DECL^U14^H TREE_CHAIN is cleared by COPY_NODE
    }
    if (value)
	value = tree_cons(NULL_TREE, value, NULL_TREE);
    TYPE_ATTRIBUTES( type ) = tree_cons( get_identifier(attrname), value,
	TYPE_ATTRIBUTES( type ));
    return type;
}

void
IRState::addDeclAttribute(tree type, const char * attrname, tree value)
{
    if (value)
	value = tree_cons(NULL_TREE, value, NULL_TREE);
    DECL_ATTRIBUTES( type ) = tree_cons( get_identifier(attrname), value,
	DECL_ATTRIBUTES( type ));
}

tree
IRState::integerConstant(xdmd_integer_t value, tree type) {
#if D_GCC_VER < 40
    // Assuming xdmd_integer_t is 64 bits
# if HOST_BITS_PER_WIDE_INT == 32
    tree tree_value = build_int_2(value & 0xffffffff, (value >> 32) & 0xffffffff);
# elif HOST_BITS_PER_WIDE_INT == 64
    tree tree_value = build_int_2(value,
	type && ! TREE_UNSIGNED(type) && (value & 0x8000000000000000ULL) ?
	~(unsigned HOST_WIDE_INT) 0 : 0);
# else
#  error Fix This
# endif
    if (type)
	TREE_TYPE( tree_value ) = type;
#else
# if HOST_BITS_PER_WIDE_INT == 32
    tree tree_value = build_int_cst_wide(type,
	value & 0xffffffff, (value >> 32) & 0xffffffff);
# elif HOST_BITS_PER_WIDE_INT == 64
    tree tree_value = build_int_cst_type(type, value);
# else
#  error Fix This
# endif
#endif
    return tree_value;
}

tree
IRState::vtable(Array * vtbl, ClassDeclaration * ci_override)
{
    tree ctor = make_node( CONSTRUCTOR );
    tree elem_list = NULL_TREE;

    TREE_TYPE( ctor ) = arrayType(Type::tvoid->pointerTo(), vtbl->dim);

    for (int i = vtbl->dim - 1; i >= 0 ; i--) { // backwards so elem_list is in order
	Dsymbol * vtbl_ent = (Declaration *) vtbl->data[i];
	Declaration * decl;
	tree value;
	
	if ((decl = vtbl_ent->isDeclaration()) && decl->isAbstract()) {
	    value = d_null_pointer;
	} else {
	    if (i == 0 && ci_override)
		vtbl_ent = ci_override;
	    value = addressOf( vtbl_ent );
	}
	
	elem_list = tree_cons(NULL_TREE, value, elem_list);
    }

    TREE_CONSTANT( ctor ) = 1;
    TREE_READONLY( ctor ) = 1;
    TREE_STATIC( ctor ) = 1;
    TREE_ADDRESSABLE( ctor ) = 1; // %% or just the decl?
    CONSTRUCTOR_ELTS( ctor ) = elem_list;
    return ctor;
}

tree
IRState::exceptionObject()
{
    tree obj_type = getObjectType()->toCtype();
    // Like gjc, the actual D exception object is one
    // pointer behind the exception header
    tree t = build (EXC_PTR_EXPR, ptr_type_node);
    t = build1(NOP_EXPR, build_pointer_type(obj_type), t); // treat exception header as ( Object* )
    t = build(MINUS_EXPR, TREE_TYPE(t), t, TYPE_SIZE_UNIT(TREE_TYPE(t)));
    t = build1(INDIRECT_REF, obj_type, t);
    return t;
}

tree
IRState::label(Loc loc, Identifier * ident) {
    tree t_label = build_decl(LABEL_DECL,
	ident ? get_identifier(ident->string) : NULL_TREE, void_type_node);
    DECL_CONTEXT( t_label ) = current_function_decl;
    DECL_MODE( t_label ) = VOIDmode;
    if (loc.filename)
	setDeclLoc(t_label, loc);
    return t_label;
}

#if D_GCC_VER < 40

void
IRState::startCond(Statement * stmt, Expression * e_cond) {
    clear_last_expr ();
    doLineNote(stmt->loc);
    expand_start_cond( convertForCondition( e_cond ), 0 );
}

void
IRState::startElse() { expand_start_else(); }

void
IRState::endCond() { expand_end_cond(); }

void
IRState::startLoop(Statement * stmt) {
    beginFlow(stmt, expand_start_loop_continue_elsewhere(1));
}

void
IRState::continueHere() { expand_loop_continue_here(); }

void
IRState::exitIfFalse(tree t_cond, bool is_top_cond) {    
    
    // %% topcond compaitble with continue_elswehre?
    expand_exit_loop_if_false(currentFlow()->loop, t_cond);
}    

void
IRState::startCase(Statement * stmt, tree t_cond)
{
    clear_last_expr ();
    doLineNote(stmt->loc);
    expand_start_case( 1, t_cond, TREE_TYPE( t_cond ), "switch statement" );
    beginFlow(stmt, NULL);
}

void
IRState::doCase(tree t_value, tree t_label)
{
    tree dummy;
    // %% not the same convert that is in d-glue!!
    pushcase( t_value, convert, t_label, & dummy);
}

void
IRState::endCase(tree t_cond)
{
    expand_end_case( t_cond );
    endFlow();
}

void
IRState::endLoop() {
    expand_end_loop();
    endFlow();
}
 
void
IRState::continueLoop(Identifier * ident) {
    expand_continue_loop( getLoopForLabel( ident, true )->loop );
}

void
IRState::exitLoop(Identifier * ident)
{
    if (ident) {
	Flow * flow = getLoopForLabel( ident );
	if (flow->loop)
	    expand_exit_loop( flow->loop );
	else {
	    if (! flow->exitLabel)
		flow->exitLabel = label(flow->statement->loc);
	    expand_goto( flow->exitLabel );
	}
    } else {
	expand_exit_something();
    }
}

void
IRState::startTry(Statement * stmt)
{
    expand_eh_region_start();
}

void
IRState::startCatches()
{
    expand_start_all_catch();
}

void
IRState::startCatch(tree t_type)
{
    expand_start_catch( t_type );
}

void
IRState::endCatch()
{
    expand_end_catch();
}

void
IRState::endCatches()
{
    expand_end_all_catch();
}

void
IRState::startFinally()
{
    abort();
}

void
IRState::endFinally()
{
    abort();
}

void
IRState::doLabel(tree t_label)
{
    expand_label(t_label);
}

void
IRState::doReturn(tree t_value)
{
    if (t_value)
	expand_return(t_value);
    else
	expand_null_return();
}

void
IRState::doJump(Statement * stmt, tree t_label)
{
    // %%TODO: c-semantics.c:expand_stmt GOTO_STMT branch prediction
    TREE_USED( t_label ) = 1 ;
    if (stmt)
	doLineNote( stmt->loc );
    expand_goto( t_label );
}

tree
IRState::makeStmtExpr(Statement * statement)
{
    tree t = build1((enum tree_code) D_STMT_EXPR, void_type_node, NULL_TREE);
    TREE_SIDE_EFFECTS(t) = 1; // %% 
    stmtExprList.push(t);
    stmtExprList.push(statement);
    stmtExprList.push(this);
    return t;
}

void
IRState::retrieveStmtExpr(tree t, Statement ** s_out, IRState ** i_out)
{
    for (int i = stmtExprList.dim - 3; i >= 0 ; i -= 3) {
	if ( (tree) stmtExprList.data[i] == t ) {
	    *s_out = (Statement *) stmtExprList.data[i + 1];
	    *i_out = (IRState *)   stmtExprList.data[i + 2];
	    // The expression could be evaluated multiples times, so we must
	    // keep the values forever --- %% go back to per-function list
	    /*
	    stmtExprList.remove(i);
	    stmtExprList.remove(i);
	    stmtExprList.remove(i);
	    */
	    return;
	}
    }
    abort();
    return;
}

#else

void
IRState::startCond(Expression * e_cond)
{
    tree t_cond = convertForCondition(e_cond);

    Flow * f = beginFlow(stmt/*, 0*/);
    f->condition = t_cond;
}

void
IRState::startElse()
{
    currentFlow()->trueBranch = popStatementList();
    pushStatementList();
}

void
IRState::endCond()
{
    Flow * f = currentFlow();
    tree t_brnch = popStatementList(); // endFlow(); -- can't pop -- need the info?
    tree t_false_brnch = NULL_TREE;

    if (f->trueBranch != NULL_TREE)
	f->trueBranch = t_brnch;
    else
	t_false_brnch = t_brnch;

    doLineNote(f->statement->loc);
    tree t_stmt = build3(COND_EXPR, void_type_node,
	f->condition, f->trueBranch, t_false_brnch);
    endFlow();
    addExpr(t_stmt);
}

void
IRState::startLoop(Statement * stmt) {
    Flow * f = beginFlow(stmt/*, Break|Continue*/);
    f->continueLabel = makeLabel(stmt ? stmt->loc : 0); // should be end for 'do' loop
}

void
IRState::continueHere() {
    doLabel(currentFlow()->continueLabel);
}

void
IRState::exitIfFalse(tree t_cond, bool unused = 0) {
    addExpr(build1(EXIT_EXPR, void_type_node, t_cond));
}    

void
IRState::endLoop() {
    // says must contain an EXIT_EXPR -- what about while(1)..goto;? something other thand LOOP_EXPR?
    tree t_body = popStatementList();
    tree t_loop = build1(LOOP_EXPR, void_type_node, t_body);
    addExpr(t_body);
    endFlow();
}

void
IRState::continueLoop(Identifier * ident) {
    //doFlowLabel(stmt, ident, Continue);
    doJump(NULL, getLoopForLabel(ident, true)->continueLoop );
}

void
IRState::exitLoop(Statement * stmt, Identifier * ident) {
    if (ident) {
	Flow * flow = getLoopForLabel(ident);
	if (! flow->exitLabel)
	    flow->exitLabel = label(flow->statement->loc);
	doJump(NULL, flow->exitLabel);
    } else
	addExpr(build1(EXIT_EXPR, void_type_node, boolean_true_node));
}


void
IRState::startTry(Statement * stmt)
{
    Flow * flow = beginFlow(stmt);
}

void
IRState::startCatches()
{
    currentFlow()->tryBody = popStatementList();
    pushStatementList();
}

void
IRState::startCatch(tree t_type)
{
    currentFlow()->catchType = t_type;
    pushStatementList();
}

void
IRState::endCatch()
{
    tree t_body = popStatementList();
    // % wrong loc.. can set pass statement to startCatch, set
    // the loc on t_type and then use it here..
    // may not be important?
    doExp(build2(CATCH_EXPR, void_type_node,
	      currentFlow()->catchType, t_body));
}

void
IRState::endCatches()
{
    tree t_catches = popStatementList();
    doLineNote(currentFlow()->statement);
    doExp(build2(TRY_CATCH_EXPR, void_type_node,
	      currentFlow()->tryBody, t_catches));
    endFlow();
}

void
IRState::startFinally()
{
    currentFlow()->tryBody = popStatementList();
    pushStatementList();
}

void
IRState::endFinally()
{
    tree t_finally = popStatementList();
    doLineNote(currentFlow()->statement);
    doExp(build2(TRY_FINALLY_EXPR, void_type_node,
	      currentFlow()->tryBody, t_finally));
    endFlow();
}
    
void
IRState::doLabel(tree t_label)
{
    addExpr(build1(LABEL_EXPR, void_type_node, t_label));
}

void
IRState::doReturn(tree t_value)
{
    addExpr(build1(RETURN_EXPR, void_type_node, t_value));
}

void
IRState::doJump(Statement * stmt, tree t_label)
{
    doLineNote( stmt->loc );
    addExpr(build1(GOTO_EXPR, void_type_node, t_label);
}

tree
IRState::makeStmtExpr(Statement * statement, IRState * irs)
{
    tree t_list;
    
    pushStatementList();
    statement->toIR(this);
    t_list = popStatementList();

    // addExpr(t_list);

    return t_list;
}

#endif


BitConfig bitConfig;

void
BitConfig::setType(Type * type)
{
    elemType = type;
    shift = exact_log2( elemType->size() * 8 );
    mask = elemType->size() * 8 - 1;
    shift_tree = IRState::integerConstant(shift, unsigned_type_node); // %% 32/64?
    mask_tree = IRState::integerConstant(mask, elemType->toCtype());
    dkeep(shift_tree);
    dkeep(mask_tree);
}

WrappedExp::WrappedExp(Loc loc, enum TOK op, tree exp_node, Type * type)
    : Expression(loc, op, sizeof(WrappedExp))
{
    this->exp_node = exp_node;
    this->type = type;
}

void
WrappedExp::toCBuffer(OutBuffer *buf)
{
    buf->printf("< wrapped exprission >");
}

elem *
WrappedExp::toElem(IRState *)
{
    return exp_node;
}


void AggLayout::doFields(Array * fields, AggregateDeclaration * agg)
{
    bool inherited = agg != this->aggDecl;
    tree fcontext;

    fcontext = agg->type->toCtype();
    if ( POINTER_TYPE_P( fcontext ))
	fcontext = TREE_TYPE(fcontext);

    // tree new_field_chain = NULL_TREE;
    for (unsigned i = 0; i < fields->dim; i++) {
	// %% D anonymous unions just put the fields into the outer struct...
	// does this cause problems?

	VarDeclaration * var_decl = (VarDeclaration *) fields->data[i];
	
	assert( var_decl->storage_class & STCfield );

	tree ident = var_decl->ident ? get_identifier(var_decl->ident->string) : NULL_TREE;
	tree field_decl = build_decl(FIELD_DECL, ident,
	    gen.trueDeclarationType( var_decl ));
	gen.setDeclLoc( field_decl, var_decl );
	var_decl->csym = new Symbol;
	var_decl->csym->Stree = field_decl;
	
	DECL_CONTEXT( field_decl ) = aggType;
	DECL_FCONTEXT( field_decl ) = fcontext;
	DECL_FIELD_OFFSET (field_decl) = size_int( var_decl->offset );
	DECL_FIELD_BIT_OFFSET (field_decl) = bitsize_zero_node;

	DECL_ARTIFICIAL( field_decl ) =
	    DECL_IGNORED_P( field_decl ) = inherited;

	//SET_DECL_OFFSET_ALIGN (field_decl, BIGGEST_ALIGNMENT); // ?
	layout_decl( field_decl, 0 );
	
	// get_inner_reference doesn't check these, leaves a variable unitialized
	// DECL_SIZE is NULL if size is zero.
	if (var_decl->size(var_decl->loc)) {
	    assert(DECL_MODE(field_decl) != VOIDmode);
	    assert(DECL_SIZE(field_decl) != NULL_TREE);
	}
	fieldList.chain( field_decl );
    }
}

void AggLayout::doInterfaces(Array * bases, AggregateDeclaration * agg)
{
    tree fcontext = TREE_TYPE( agg->type->toCtype() );
    for (unsigned i = 0; i < bases->dim; i++) {
	BaseClass * bc = (BaseClass *) bases->data[i];
	tree decl = build_decl(FIELD_DECL, NULL_TREE,
	    Type::tvoid->pointerTo()->pointerTo()->toCtype() /* %% better */ );
	//DECL_VIRTUAL_P( decl ) = 1; %% nobody cares, boo hoo
	DECL_ARTIFICIAL( decl ) =
	    DECL_IGNORED_P( decl ) = 1;
	// DECL_FCONTEXT( decl ) = fcontext; // shouldn't be needed since it's ignored
	addField(decl, bc->offset);
    }
}

void AggLayout::addField(tree field_decl, unsigned offset)
{
    DECL_CONTEXT( field_decl ) = aggType;
    // DECL_FCONTEXT( field_decl ) = aggType; // caller needs to set this
    DECL_FIELD_OFFSET (field_decl) = size_int( offset );
    DECL_FIELD_BIT_OFFSET (field_decl) = bitsize_int( 0 );
    Loc l(aggDecl->getModule(), 1); // Must set this or we crash with DWARF debugging
    // gen.setDeclLoc( field_decl, aggDecl->loc ); // aggDecl->loc isn't set
    gen.setDeclLoc(field_decl, l);
    
    layout_decl( field_decl, 0 );
    fieldList.chain( field_decl );
}

void AggLayout::finish()
{
    unsigned size_to_use = aggDecl->structsize;
    unsigned align_to_use = aggDecl->alignsize;

    /* probably doesn't do anything */
    /*
    if (aggDecl->structsize == 0 && aggDecl->isInterfaceDeclaration())
	size_to_use = Type::tvoid->pointerTo()->size();
    */

    TYPE_SIZE( aggType ) = bitsize_int( size_to_use * BITS_PER_UNIT );
    TYPE_SIZE_UNIT( aggType ) = size_int( size_to_use );
    TYPE_ALIGN( aggType ) = align_to_use * BITS_PER_UNIT; // %%doc int, not a tree
    // TYPE_ALIGN_UNIT is not an lvalue
    TYPE_PACKED (aggType ) = TYPE_PACKED (aggType ); // %% todo
    compute_record_mode ( aggType );
    // %%  stor-layout.c:finalize_type_size ... it's private to that file

    // c-decl.c -- set up variants ? %%
    for (tree x = TYPE_MAIN_VARIANT( aggType ); x; x = TYPE_NEXT_VARIANT( x )) {
	TYPE_FIELDS (x) = TYPE_FIELDS (aggType);
	TYPE_LANG_SPECIFIC (x) = TYPE_LANG_SPECIFIC (aggType);
	TYPE_ALIGN (x) = TYPE_ALIGN (aggType);
	TYPE_USER_ALIGN (x) = TYPE_USER_ALIGN (aggType);
    }
}

ArrayMaker::ArrayMaker(Type * elem_type, int storage_class)
{
    elemType = elem_type;
    storageClass = storage_class;
    
    elemList = NULL_TREE;
    count = 0;
}

ArrayMaker &
ArrayMaker::add(tree value)
{
    elemList = tree_cons(NULL_TREE, value, elemList);
    count++;
    return *this;
}

tree
ArrayMaker::finish()
{
    tree ctor = make_node(CONSTRUCTOR);
    tree type = gen.arrayType(elemType->toCtype(), count);
    TREE_TYPE( ctor ) = type;
    CONSTRUCTOR_ELTS( ctor ) = nreverse( elemList );
    if (storageClass & STCstatic)
	TREE_STATIC( ctor ) = 1;
    if (storageClass & STCconst)
	TREE_CONSTANT( ctor ) = 1;
    TREE_READONLY( ctor ) = 1;
    return ctor;
    
}

ArrayScope::ArrayScope(IRState * ini_irs, VarDeclaration * ini_v, const Loc & loc) :
    v(ini_v), irs(ini_irs)
{
    if (v) {
	/* Need to set the location or the expand_decl in the BIND_EXPR will
	   cause the line numbering for the statement to be incorrect. */
	/* The variable itself is not included in the debugging information. */
	v->loc = loc; 
	Symbol * s = v->toSymbol();
	tree decl = s->Stree;
	DECL_CONTEXT( decl ) = irs->getCurrentFunction()->toSymbol()->Stree;
    }
}

tree
ArrayScope::setArrayExp(tree e, Type * t)
{
    if ( v) {
	if (t->toBasetype()->ty != Tsarray)  // %% 
	    e = irs->maybeMakeTemp(e);
	DECL_INITIAL(v->toSymbol()->Stree) = irs->arrayLength(e, t);
    }
    return e;
}

tree
ArrayScope::finish(tree e)
{
    if (v) {
	return build(BIND_EXPR, TREE_TYPE(e), v->toSymbol()->Stree,
	    e, NULL_TREE);
    } else {
	return e;
    }
}

void
FieldVisitor::visit(AggregateDeclaration * decl)
{
    ClassDeclaration * class_decl = decl->isClassDeclaration();
    
    if (class_decl && class_decl->baseClass)
	FieldVisitor::visit(class_decl->baseClass);

    doFields(& decl->fields, decl);

    if (class_decl)
	doInterfaces(class_decl->vtblInterfaces, decl);
}

dt_t** dtnwords(dt_t** pdt, unsigned word_count, void * pwords, unsigned word_size)
{
    return dtnbytes(pdt, word_count * word_size,
	gen.hostToTargetString((char*) pwords, word_count, word_size));
}

dt_t** dtawords(dt_t** pdt, unsigned word_count, void * pwords, unsigned word_size)
{
    return dtabytes(pdt, TYnptr, 0, word_count * word_size,
	gen.hostToTargetString((char*) pwords, word_count, word_size));
}

unsigned dt_size(dt_t * dt)
{
    unsigned size = 0;

    while (dt) {
	switch (dt->dt) {
	    case DT_azeros:
	    case DT_common:
	    case DT_nbytes:	
		size += dt->DTint;
		break;
	    case DT_abytes:
	    case DT_xoff:
		size += PTRSIZE;
		break;
	    case DT_word:
		// This may cause problems for 64-bit since dtdword is used for
		// (uint? / size_t?) and pointers... Nearly all cases are for size_t
		// and pointer, though.
		size += PTRSIZE;
		break;
	    case DT_tree:
		size += tree_low_cst( TYPE_SIZE_UNIT( TREE_TYPE( dt->DTtree )), 1 );
		break;
	}
	dt = dt->DTnext;
    }

    return size;
}

static tree dt2node(dt_t * dt)
{
    tree a;
    switch (dt->dt) {
    case DT_azeros:
    case DT_common:
	a = make_node(CONSTRUCTOR);
	TREE_TYPE(a) = gen.arrayType(Type::tuns8, dt->DTint);
	return a;
    case DT_nbytes:
	{
	    tree s = build_string(dt->DTint, (char *) dt->DTpointer);
	    TREE_TYPE( s ) = gen.arrayType(Type::tuns8, dt->DTint);	    
	    return s;
	}
    case DT_abytes:
	{
	    tree s = build_string(dt->DTint, (char *) dt->DTpointer);
	    TREE_TYPE( s ) = gen.arrayType(Type::tuns8, dt->DTint);
	    TREE_STATIC( s ) = 1;
	    return gen.addressOf( s );
	}
    case DT_word:
	// %% make sure this is the target word type
	return gen.integerConstant(dt->DTint, long_unsigned_type_node);
    case DT_xoff:
	if ( (dt->DTsym->Sfl & FLstatic_sym) && ! dt->DTsym->Stree ) {
	    dt->DTsym->Stree = dt2tree( dt->DTsym->Sdt );
	}
	return build(PLUS_EXPR, ptr_type_node, gen.addressOf(dt->DTsym->Stree),
	    gen.integerConstant(dt->DTint, long_unsigned_type_node) );
    case DT_tree:
	return dt->DTtree;
    default:
	abort();
    }
    return NULL;
}

static tree dt2tree(dt_t * dt)
{
    ListMaker elts;
    ListMaker fields;
    tree offset = size_zero_node;

    tree aggtype = make_node(RECORD_TYPE);

    while (dt) {
	tree value = dt2node(dt);
	tree field = build_decl(FIELD_DECL, NULL_TREE, TREE_TYPE(value));
	DECL_CONTEXT(field) = aggtype;
	DECL_FIELD_OFFSET(field) = offset;
	DECL_FIELD_BIT_OFFSET(field) = bitsize_zero_node;
	DECL_ARTIFICIAL(field) = 1;
	DECL_IGNORED_P(field) = 1;
	layout_decl(field, 0);
	
	fields.chain(field);
	elts.cons(field, value);
	offset = size_binop(PLUS_EXPR, offset, TYPE_SIZE_UNIT( TREE_TYPE( value )));

	dt = dt->DTnext;
    }
    
    TYPE_FIELDS(aggtype) = fields.head; // or finish_laout
    TYPE_SIZE(aggtype) = convert( bitsizetype,
	size_binop( MULT_EXPR, offset, size_int( BITS_PER_UNIT )));
    TYPE_SIZE_UNIT(aggtype) = offset;
    // okay no alignment -- decl (which has the correct type) should take care of it..
    // align=bits per word?
    compute_record_mode(aggtype);

    tree ctor = make_node(CONSTRUCTOR);
    TREE_TYPE(ctor) = aggtype;
    TREE_READONLY(ctor) = 1;
    // dt created data is always static
    TREE_STATIC(ctor) = 1;
    CONSTRUCTOR_ELTS(ctor) = elts.head;
    return ctor;    
}

void outdata(Symbol * sym)
{
    if (sym->Sfl & FLstatic_sym)
	// This is a CONSTRUCTOR that will be output automatically
	return;
    
    assert(sym->Stree);

    tree t = sym->Stree;

    if ( DECL_EXTERNAL( t ) ) {
	// %% Oops, this was supposed to be static.
	// This is for typeinfo decls ... may
	// insert into module instead?
	DECL_EXTERNAL( t ) = 0;
	TREE_STATIC( t ) = 1;
    }
    
    if (sym->Sdt && DECL_INITIAL( t ) == NULL_TREE) {
	tree ini = dt2tree( sym->Sdt );
	TREE_CONSTANT( ini ) = TREE_CONSTANT( t );
	DECL_INITIAL( t ) = ini;
    }
    
    if (! gen.shouldEmit(sym))
	return;

    if (sym->Sclass == SCcomdat) {
	D_DECL_ONE_ONLY(t) = 1;
    }
    
    gen.prepareSymbolOutput(sym);
    layout_decl(t, 0);

    // from outputStaticSymbol:
    if ( DECL_SIZE(t) == NULL_TREE ) {
	DECL_SIZE(t) = bitsize_int(0);
	DECL_SIZE_UNIT(t) = size_int(0);
    } // Otherwise, if DECL_SIZE == 0, just let it fail...

    d_add_global_function(t);
    
    IRState::rodc(t, 1);
}

void obj_includelib(char *name)
{
    ::warning("pragma(lib) not implemented");
}
